class GoogleDtype(type):
    """
    Metaclass for the four data types used in Google Sheets.
    """

    python_type: type
    type_key: str

    def __str__(cls) -> str:
        return cls.type_key

    @classmethod
    def parse(cls, value: str) -> str:
        return cls.python_type(value)



from __future__ import annotations

from typing import Any, Dict, List, Literal
from pathlib import Path
import mimetypes
from warnings import warn

from googleapiclient.http import MediaFileUpload

from . import _google_terms as terms
from ._conn import Connection
from .dtypes import EffectiveFmt, EffectiveVal, FormattedVal, UserEnteredVal
from .interfaces import AuthConfig


class FileUpload:
    """
    Use FileUploads when you need more complicated file upload instructions.
    """

    def __init__(
        self,
        path: Path | str,
        to_folder_id: str | None = None,
        convert: bool | None = False,
    ) -> None:
        """

        Args:
            path (Path, str): The pathlike to the file.
            to_folder_id (str, optional): A folder id, if you want to upload this
                file to a specific folder in Google Drive. Defaults to None.
            convert (bool, optional): Set to True if you want to convert the
                file to a Google Drive format. The format will be automatically
                determined based on Google Drive's rules for conversion. Defaults
                to False.
        """
        self.path = Path(path)
        self.folder = to_folder_id
        self._do_conv = True if convert == True else False

    @property
    def do_conv(self) -> bool:
        return self._do_conv


class DriveConnection(Connection):
    """
    Provides a connection to the Google Drive api, and methods to send requests
    to it.

    This class is documented primarily for reference, if you want to connect to
    Google Drive, it's easier to just use a :class:`Drive <autodrive.drive.Drive>`
    instance.
    """

    def __init__(
        self,
        *,
        auth_config: AuthConfig | None = None,
        api_version: str = "v3",
    ) -> None:
        """

        Args:
            auth_config (AuthConfig, optional): Optional custom AuthConfig object,
                defaults to None.
            api_version (str, optional): The version of the Drive api to connect to,
                defaults to "v3".

        """
        super().__init__(
            api_name="drive",
            api_version=api_version,
            auth_config=auth_config,
        )
        self._files = self._core.files()  # type: ignore
        self._fmt_map = self.get_import_formats()

    def get_import_formats(self) -> Dict[str, str]:
        """
        Returns:
            Dict[str, str]: The mapping between standard MIMEtypes and the
                MIMEtypes for Google Drive's types (Docs, Sheets, etc) as
                provided by the Google Drive API.
        """
        resp: Dict[str, Dict[str, List[str]]] = (
            self._core.about().get(fields="importFormats").execute()  # type: ignore
        )
        result: Dict[str, str] = {}
        for mtype, google_mtypes in resp["importFormats"].items():
            if len(google_mtypes) > 1:
                warn(
                    f"Google MIMEtype mapping for {mtype} has "
                    f"{len(google_mtypes)} possible types. Please submit an bug "
                    "report at https://github.com/chrislarabee/autodrive."
                )
            result[mtype] = google_mtypes[0]
        return result

    def find_object(
        self,
        obj_name: str,
        obj_type: Literal["sheet", "folder", "file"] | None = None,
        shared_drive_id: str | None = None,
    ) -> List[Dict[str, Any]]:
        """
        Searches for a Google Drive Object via the connected api.

        Args:
            obj_name (str): The name of the object, or part of its name.
            obj_type (Literal["sheet", "folder", "file"], optional): The type of
                object to restrict the search to.
            shared_drive_id (str, optional): The id of a Shared Drive to search
                within, if desired, defaults to None.

        Returns:
            List[Dict[str, Any]]: A list of object properties, if any matches are
            found.

        """
        query = f"name = '{obj_name}'"
        if obj_type == "file":
            query += f" and mimeType != '{self.google_obj_types['folder']}'"
        elif obj_type:
            query += f" and mimeType='{self.google_obj_types[obj_type]}'"
        kwargs = self._setup_drive_id_kwargs(shared_drive_id)
        page_token = None
        results: List[Dict[str, Any]] = []
        while True:
            response = self._files.list(  # type: ignore
                q=query,
                spaces="drive",
                fields=f"nextPageToken, files ({terms.ID},{terms.NAME},{terms.PARENTS})",
                pageToken=page_token,
                **kwargs,
            ).execute()
            for file in response.get("files", []):  # type: ignore
                results.append(
                    dict(
                        name=file.get(terms.NAME),  # type: ignore
                        id=file.get(terms.ID),  # type: ignore
                        parents=file.get(terms.PARENTS),  # type: ignore
                    )
                )
            page_token = response.get("nextPageToken", None)  # type: ignore
            if page_token is None:
                break
        return results

    def create_object(
        self,
        obj_name: str,
        obj_type: Literal["sheet", "folder"],
        parent_id: str | None = None,
    ) -> str:
        """
        Creates a file or folder via the Google Drive connection.

        Args:
            obj_name (str): The desired name of the object to create.
            obj_type (Literal["sheet", "folder"]): The type of object to create.
            parent_id (str, optional): The id of the folder or shared drive to
                create the object within, defaults to None.

        Returns:
            str: The new id of the created object.

        """
        kwargs: Dict[str, List[str]] = (
            dict(parents=[parent_id]) if parent_id else dict()
        )
        file_metadata: Dict[str, Any] = dict(
            name=obj_name, mimeType=self.google_obj_types[obj_type], **kwargs
        )
        file = self._files.create(  # type: ignore
            body=file_metadata, fields=terms.ID, supportsAllDrives=True
        ).execute()
        return file.get(terms.ID)  # type: ignore

    def delete_object(self, object_id: str) -> None:
        """
        Deletes the passed Google object id from the connected Google Drive.

        Args:
            object_id (str): A Google object id.

        """
        self._files.delete(  # type: ignore
            fileId=object_id, supportsAllDrives=True
        ).execute()

    def upload_files(self, *filepaths: Path | str | FileUpload) -> Dict[str, str]:
        """
        Uploads files to the root drive or to a folder.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            *filepaths (Path | str | FileUpload): An arbitrary number of Path
                objects, path strings, or FileUpload objects (for more complex
                cases).

        Returns:
            Dict[str, str]: The names of the uploaded files and their new ids in
                Google Drive.
        """
        result: Dict[str, str] = {}
        kwargs: Dict[str, Any] = {}
        for fp in filepaths:
            if isinstance(fp, FileUpload):
                path = fp.path
                if fp.folder:
                    kwargs["parents"] = [fp.folder]
                if fp.do_conv:
                    kwargs["mimeType"] = self.detect_conv_format(fp.path)
            else:
                path = Path(fp)
            file_metadata: Dict[str, str] = dict(name=path.name, **kwargs)
            mtype, _ = mimetypes.guess_type(path)
            media = MediaFileUpload(path, mimetype=mtype)
            resp = self._files.create(  # type: ignore
                body=file_metadata,
                media_body=media,
                fields=terms.ID,
            ).execute()
            id: str = resp.get(terms.ID)  # type: ignore
            result[path.name] = id
        return result

    def detect_conv_format(self, p: Path) -> str:
        """
        Detects the MIMEtype for the passed filepath and determines its
        corresponding Google Drive format.

        Args:
            p (Path): Any filepath.

        Raises:
            ValueError: If the MIMEtype cannot be determined from the path at
                all.
            ValueError: If the MIMEtype cannot be converted to a Google Drive
                format MIMEtype.

        Returns:
            str: The Google Drive format MIMEtype for the file.
        """
        mtype, _ = mimetypes.guess_type(p)
        if mtype is None:
            raise ValueError(
                f"Could not determine MIMEtype for {p}. Unable to convert to "
                "Google Drive format."
            )
        elif mtype in self._fmt_map.keys():
            return self._fmt_map[mtype]
        else:
            raise ValueError(
                f"File {p} has MIMEtype {mtype} which is not valid for "
                "conversion to Google Drive format."
            )

    @staticmethod
    def _setup_drive_id_kwargs(drive_id: str | None = None) -> Dict[str, str | bool]:
        """
        Whenever a drive_id is needed to access a shared drive, two other kwargs
        need to be passed to the relevant function. This method preps all three
        kwargs.

        Args:
            drive_id (str, optional): The id of the shared drive, defaults to None.

        Returns:
            Dict[str, str | bool]: A dictionary containing the kwargs needed to
            access the shared drive.

        """
        kwargs: Dict[str, str | bool] = dict()
        if drive_id:
            kwargs["corpora"] = "drive"
            kwargs["driveId"] = drive_id
            kwargs["includeItemsFromAllDrives"] = True
            kwargs["supportsAllDrives"] = True
        return kwargs


class SheetsConnection(Connection):
    """
    Provides a connection to the Google Sheets api, and methods to send requests
    to it.

    This class is documented primarily for reference, if you want to connect to
    Google Sheets, it's easier to just use a View.
    """

    def __init__(
        self,
        *,
        auth_config: AuthConfig | None = None,
        api_version: str = "v4",
    ) -> None:
        """
        Args:
            auth_config (AuthConfig, optional): Optional custom AuthConfig object,
                defaults to None.
            api_version (str, optional): The version of the Sheets api to connect
                to, defaults to "v4".

        """
        super().__init__(
            api_name="sheets",
            api_version=api_version,
            auth_config=auth_config,
        )
        self._sheets = self._core.spreadsheets()  # type: ignore

    def execute_requests(
        self, spreadsheet_id: str, requests: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        """
        Sends the passed list of request dictionaries to the Sheets api to be
        applied to the spreadsheet_id via batch update.

        Args:
            spreadsheet_id (str): The id of the Google Sheet to update.
            requests (List[Dict[str, Any]]): A list of dictionaries formatted as
                requests.

        Returns:
            Dict[str, Any]: The resulting response from the Sheets api as a
            dictionary.

        """
        result: Dict[str, Any] = self._sheets.batchUpdate(  # type: ignore
            spreadsheetId=spreadsheet_id, body=self._preprocess_requests(requests)
        ).execute()
        return result

    def get_properties(self, spreadsheet_id: str) -> Dict[str, Any]:
        """
        Gets the metadata properties of the indicated Google Sheet.

        Args:
            spreadsheet_id (str): The id of the Google Sheet to collect properties
                from.

        Returns:
            Dict[str, Any]: A dictionary of the Google Sheet's properties.

        """
        gsheet_props = f"{terms.FILE_PROPS}({terms.FILE_NAME})"
        grid_props = f"{terms.GRID_PROPS}({terms.COL_CT},{terms.ROW_CT})"
        tab_props = f"{terms.TAB_IDX},{terms.TAB_ID},{terms.TAB_NAME},{grid_props}"
        tabs_prop = f"{terms.TABS_PROP}({terms.TAB_PROPS}({tab_props}))"
        return self._sheets.get(  # type: ignore
            spreadsheetId=spreadsheet_id, fields=f"{gsheet_props},{tabs_prop}"
        ).execute()

    def get_data(
        self, spreadsheet_id: str, ranges: List[str] | None = None
    ) -> Dict[str, Any]:
        """
        Collects data from cells in the passed spreadsheet.

        Args:
            spreadsheet_id (str): The id of the Google Sheet to collect data from.
            ranges (List[str], optional): A list of range strings
                (e.g. Sheet1!A1:C3), defaults to None, which prompts get_data to
                fetch all data from all cells.

        Returns:
            Dict[str, Any]: The collected data from the spreadsheet, raw and
            unparsed.

        """
        data_values = f"{UserEnteredVal},{FormattedVal},{EffectiveVal}"
        formatting_values = f"{EffectiveFmt}"
        values = f"{terms.VALUES}({data_values},{formatting_values})"
        return self._sheets.get(  # type: ignore
            spreadsheetId=spreadsheet_id,
            fields=f"{terms.TABS_PROP}({terms.DATA}({terms.ROWDATA}({values})))",
            ranges=ranges or [],
        ).execute()



from __future__ import annotations

from typing import List

from .._view import CellFormatting, GridFormatting, TextFormatting
from ..interfaces import BorderFormat, Color, Format, HalfRange, FullRange
from . import _cell as cell, _grid as grid, _text as text
from ..dtypes import (
    BorderSide,
    BorderStyle,
    BorderSides,
    HorizontalAlign,
    VerticalAlign,
)


class TabCellFormatting(CellFormatting):
    """
    Contains request generation methods relating to formatting this Tab's cells
    (like adding borders and backgrounds and such).
    """

    def add_alternating_row_background(
        self, colors: Color, rng: FullRange | str | None = None
    ) -> TabCellFormatting:
        """
        Queues a request to add an alternating row background of the indicated
        color to this Tab's cells, or to a range within the Tab.

        Args:
            colors (Color): The desired Color to apply to every other row.
            rng (FullRange | str, optional): Optional range within the Tab to
                apply the format to, defaults to None, for all cells in the Tab.

        Returns:
            TabCellFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        rng = self.ensure_full_range(rng)
        self.add_request(
            cell.add_alternating_row_background(self._parent.tab_id, rng, colors)
        )
        return self

    def set_background_color(
        self, color: Color, rng: FullRange | str | None = None
    ) -> TabCellFormatting:
        """
        Queues a request to set the background of the Tab's cells (or the
        specified cells within the Tab) to the indicated color

        Args:
            color (Color): The desired Color to set the background to.
            rng (FullRange | str, optional): Optional range within the Tab to
                apply the format to, defaults to None, for all cells in the Tab.

        Returns:
            TabCellFormatting: This formatting object, so further requests can be
            queued if desired.
        """
        rng = self.ensure_full_range(rng)
        self.add_request(cell.set_background_color(self._parent.tab_id, rng, color))
        return self

    def set_border_format(
        self,
        *sides: BorderSide,
        style: BorderStyle | None = None,
        color: Color | None = None,
        rng: FullRange | str | None = None
    ) -> TabCellFormatting:
        """
        Queues a request to set the border properties of the Tab's cells (or the
        specified cells within the Tab).

        Args:
            *sides: (BorderSide): One or more BorderSide objects, indicating
                which side(s) of the cells you want to apply the border
                properties to. If no sides are provided, set_border_format will
                apply border properties to all sides.
            style (BorderStyle, optional): The style to apply to all the
                indicated sides. Defaults to None, for the default border style.
            color (Color, optional): The color to set the border(s) to. Defaults
                to None, for black.
            rng (FullRange | str, optional): Optional range within the Tab to
                apply the format to, defaults to None, for all cells in the Tab.

        Returns:
            TabCellFormatting: This formatting object, so further requests can be
            queued if desired.
        """
        rng = self.ensure_full_range(rng)
        fmts: List[BorderFormat] = []
        if len(sides) == 0:
            sides = BorderSides
        for side in sides:
            fmts.append(BorderFormat(side, color, style))
        self.add_request(cell.set_border_format(self._parent.tab_id, rng, *fmts))
        return self


class TabGridFormatting(GridFormatting):
    """
    Contains request generation methods related to formatting this Tab's grid
    (number of columns, rows, width and height, etc).
    """

    def auto_column_width(self, rng: HalfRange | None = None) -> TabGridFormatting:
        """
        Queues a request to set the column width of the Tab's columns equal to the
        width of the values in the cells.

        Args:
            rng (HalfRange, optional): The range of columns to be affected,
                defaults to None for all columns in the Tab.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        rng = rng if rng else self._parent.range.col_range
        self.add_request(grid.auto_column_width(self._parent.tab_id, rng))
        return self

    def append_rows(self, num_rows: int) -> TabGridFormatting:
        """
        Queues a request to add new empty rows at the bottom of the Tab.

        Args:
            num_rows (int): The number of rows to add to the bottom of the Tab.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
             queued if desired.

        """
        self.add_request(grid.append_rows(self._parent.tab_id, num_rows))
        return self

    def insert_rows(self, num_rows: int, at_row: int) -> TabGridFormatting:
        """
        Queues a request to insert new empty rows at the specified row number.

        Args:
          num_rows (int): The number of rows to insert.
          at_row (int): The row number to insert after.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        self.add_request(grid.insert_rows(self._parent.tab_id, num_rows, at_row - 1))
        return self

    def delete_rows(self, rng: HalfRange) -> TabGridFormatting:
        """
        Queues a request to delete rows in the selected row range.

        Args:
          rng (HalfRange): The range of rows to delete.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        self.add_request(grid.delete_rows(self._parent.tab_id, rng))
        return self

    def append_columns(self, num_cols: int) -> TabGridFormatting:
        """
        Queues a request to add new empty columns at the right of the Tab.

        Args:
            num_cols (int): The number of columns to add to the right of the Tab.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        self.add_request(grid.append_columns(self._parent.tab_id, num_cols))
        return self

    def insert_columns(self, num_cols: int, at_col: int) -> TabGridFormatting:
        """
        Queues a request to insert new empty columns at the specified column
        number.

        Args:
          num_cols (int): The number of columns to insert.
          at_col (int): The column number to insert after.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        self.add_request(grid.insert_columns(self._parent.tab_id, num_cols, at_col))
        return self

    def delete_columns(self, rng: HalfRange) -> TabGridFormatting:
        """
        Queues a request to delete columns in the selected column range.

        Args:
          rng (HalfRange): The range of columns to delete.

        Returns:
            TabGridFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        self.add_request(grid.delete_columns(self._parent.tab_id, rng))
        return self


class TabTextFormatting(TextFormatting):
    """
    Contains request generation methods relating to formatting this Tab's text
    (the text format of any cells, even those containing non-text values like
    integers or null values).
    """

    def apply_format(
        self, format: Format, rng: FullRange | str | None = None
    ) -> TabTextFormatting:
        """
        Queues a request to set the text/number format of the Tab's cells (or the
        specified cells within the Tab).

        Args:
            format (Format): A format instance, such as TextFormat or NumberFormat.
            rng (FullRange | str, optional): Optional range within the Tab to
                apply the format to, defaults to None, for all cells in the Tab.

        Returns:
            TabTextFormatting: This formatting object, so further requests can be
            queued if desired.

        """
        rng = self.ensure_full_range(rng)
        self.add_request(text.apply_format(self._parent.tab_id, rng, format))
        return self

    def set_alignment(
        self,
        *aligns: HorizontalAlign | VerticalAlign,
        rng: FullRange | str | None = None
    ) -> TabTextFormatting:
        """
        Queues a request to set the horizontal and/or vertical text alignment of
        the Tab's cells (or the specified cells within the Tab).

        Args:
            aligns (HorizontalAlign | VerticalAlign): The desired horizontal
                and/or vertical alignment properties. Note that if you specify
                a HorizontalAlign more than once, or a VerticalAlign more than
                once, only the last of each will be used.
            rng (FullRange | str, optional): Optional range within the Tab to
                apply the format to, defaults to None, for all cells in the Tab.

        Returns:
            TabTextFormatting: This formatting object, so further requests can be
            queued if desired.
        """
        rng = self.ensure_full_range(rng)
        halign = None
        valign = None
        for align in aligns:
            if isinstance(align, HorizontalAlign):
                halign = align
            else:
                valign = align
        self.add_request(
            text.set_text_alignment(self._parent.tab_id, rng, halign, valign)
        )
        return self



from __future__ import annotations

from functools import singledispatchmethod
from typing import Any, Dict, KeysView, List, Optional, ValuesView, Sequence, Literal
from pathlib import Path

from .connection import SheetsConnection
from ._view import GSheetView
from .interfaces import AuthConfig, FullRange
from .tab import Tab
from .range import Range
from .dtypes import EffectiveVal, GoogleValueType


class GSheet(GSheetView):
    """
    Provides a connection to a single Google Sheet and access to its properties
    and Tabs.
    """

    def __init__(
        self,
        gsheet_id: str,
        title: str | None = None,
        *,
        tabs: List[Tab] | None = None,
        auth_config: AuthConfig | None = None,
        sheets_conn: SheetsConnection | None = None,
        autoconnect: bool = True,
    ) -> None:
        """
        Args:
            gsheet_id (str): The id string of the target Google Sheet; can be found
                in the link to the Google Sheet.
            title (str, optional): The name of the Google Sheet, defaults to None.
            tabs (List[Tab], optional): A list of Tabs attached to the Google Sheet.
                You should probably manage Tabs with GSheet.fetch () or by getting
                the GSheet directly from a Drive, rather than using this parameter,
                defaults to None.
            auth_config (AuthConfig, optional): Optional custom AuthConfig object,
                defaults to None.
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.
            autoconnect (bool, optional): If you want to instantiate a GSheet
                without immediately checking your authentication credentials and
                connecting to the Google Sheets api, set this to False, defaults to
                True.

        """
        super().__init__(
            gsheet_id=gsheet_id,
            auth_config=auth_config,
            sheets_conn=sheets_conn,
            autoconnect=autoconnect,
        )
        self._title = title
        self._tabs = tabs or []

    @property
    def requests(self) -> List[Dict[str, Any]]:
        """
        List of accumulated (uncommitted) requests on this GSheet.

        Returns:
            List[Dict[str, Any]]: List of update request dictionaries that have
            been created for this GSheet.

        """
        return self._requests

    @property
    def tabs(self) -> Dict[str, Tab]:
        """
        Dictionary of fetched Tabs on this GSheet by title.

        Returns:
            Dict[str, Tab]: Tab titles as keys and corresponding Tabs as values.

        """
        return {tab.title: tab for tab in self._tabs}

    @property
    def title(self) -> Optional[str]:
        """
        The name of the GSheet.

        Returns:
            Optional[str]: The name of the GSheet, or None if its name hasn't been
            fetched.

        """
        return self._title

    def fetch(self) -> GSheet:
        """
        Gets the latest metadata from the API for this GSheet. Populates title and
        tab properties.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Returns:
          GSheet: This GSheet

        """
        properties = self.conn.get_properties(self._gsheet_id)
        name, sheets = self._parse_properties(properties)
        self._title = name
        tabs = [
            Tab.from_properties(self._gsheet_id, props, sheets_conn=self._conn)
            for props in sheets
        ]
        self._tabs = tabs
        return self

    @singledispatchmethod
    def add_tab(self, tab: Tab) -> GSheet:
        """
        Adds a Tab to the GSheet.

        Args:
            tab (Tab): The Tab instance you want to add.

        Returns:
            GSheet: This Gsheet.

        Raises:
            ValueError: If the GSheet already has a Tab with that title.

        """
        if tab.title in self.tabs.keys():
            raise ValueError(f"GSheet already has tab with title {tab.title}")
        self._tabs.insert(tab.index, tab)
        self._requests.append(tab.gen_add_tab_request())
        return self

    def gen_range(self, rng: FullRange, tab: str | int | None = None) -> Range:
        """
        Convenience method for generating a new Range object from a Tab in this
        GSheet.

        Args:
            rng (FullRange): The desired FullRange of the new Range object.
            tab (str | int, optional): The name of the Tab to generate from, or
                its index. Defaults to None, which will generate a range from the
                first Tab.

        Returns:
            Range: The newly generated Range object.
        """
        if isinstance(tab, str):
            from_tab = self.tabs[tab]
        elif isinstance(tab, int):
            from_tab = self._tabs[tab]
        else:
            from_tab = self._tabs[0]
        return from_tab.gen_range(rng)

    # @add_tab.register
    # def add_tab_by_name(
    #     self,
    #     tab: str,
    #     tab_id: int = None,
    #     tab_idx: int = None,
    #     num_rows: int = 1000,
    #     num_cols: int = 26,
    # ) -> GSheet:
    #     if tab in self.tabs.keys():
    #         raise ValueError(f"GSheet already has tab with title {tab}")
    #     req = Tab.new_tab_request(
    #         tab, tab_id=tab_id, tab_idx=tab_idx, num_rows=num_rows, num_cols=num_cols
    #     )
    #     self._requests.append(req)
    #     return self

    def write_values(
        self,
        data: Sequence[Sequence[Any] | Dict[str, Any]],
        to_tab: str | None = None,
        rng: FullRange | str | None = None,
        mode: Literal["write", "w", "append", "a"] = "write",
    ) -> GSheet:
        """
        Adds a request to write data. GSheet.commit () to commit the requests.

        Args:
            data (Sequence[Sequence[Any] | Dict[str, Any]]): The data to write.
                Each sequence or dictionary in the passed data is a row, with
                each value in that sub-iterable being a column. Dictionary keys
                will be used as a header row in the written data.
            to_tab (str, optional): The name of the tab to write to, defaults to
                None, which will write to whatever tab is first in the Sheet.
            rng (FullRange | str, optional): The range to which the data will be
                written, starting with the top-left-most cell in the range,
                defaults to None, which will write to the top-left-most cell in
                the passed tab, or the first tab.
            mode (Literal, optional): Whether to append the data after any
                populated rows already present in the tab or to write to the
                passed rng. Overrides rng if specified. Defaults to "write".

        Returns:
            GSheet: This GSheet.

        Raises:
            KeyError: If the passed tab name (to_tab) isn't present in the GSheet's
                current tabs property.

        """
        tab = self.tabs.get(to_tab) if to_tab else self._tabs[0]
        if not tab:
            raise KeyError(f"{to_tab} not found in {self._title} tabs.")
        # TODO: Could remove all this if GSheet can commit requests on its child
        #       Tabs...
        rng = self.ensure_full_range(tab.full_range(), rng)
        rng_dict = rng.to_dict() if mode in ["write", "w"] else None
        self._write_values(data, tab._tab_id, rng_dict)
        return self

    def get_data(
        self,
        tab: str | int | None = None,
        rng: FullRange | str | None = None,
        value_type: GoogleValueType = EffectiveVal,
    ) -> GSheet:
        """
        Gets the data from the cells of the GSheet.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            tab (str | int, optional): The name of the tab, or its (0-based) index
                (from left to right), defaults to None, which will collect data from
                the first tab in the Sheet.
            rng (FullRange | str, optional): The specific range to fetch data
                from, defaults to None, for all data in the target tab.
            value_type (GoogleValueType, optional): Allows you to toggle the
                type of the values returned by the Google Sheets API. See the
                :mod:`dtypes <autodrive.dtypes>` documentation for more info on
                the different GoogleValueTypes.

        Returns:
            GSheet: This GSheet.

        Raises:
            KeyError: If the passed tab name is not found in this GSheet's tabs.
            TypeError: If anything other than the displayed types is passed for
                the tab parameter.

        """
        if isinstance(tab, str):
            tab_ = self.tabs.get(tab)
            if not tab_:
                raise KeyError(f"{tab} not found in GSheet tabs.")
        elif isinstance(tab, int) or tab is None:
            tab_ = self._tabs[tab or 0]
        else:
            raise TypeError(
                f"tab must be a string, integer, or None. type = {type(tab)}"
            )
        rng = self.ensure_full_range(tab_.full_range(), rng)
        if not rng.tab_title:
            rng.tab_title = tab_.title
        values, formats = self._get_data(self._gsheet_id, str(rng), value_type)
        tab_.values = values
        tab_.formats = formats
        return self

    def __iter__(self):
        return self._tabs

    def __len__(self):
        return len(self._tabs)

    def __getitem__(self, key: int | str) -> Tab:
        if isinstance(key, int):
            return self._tabs[key]
        else:
            return self.tabs[key]

    def keys(self) -> KeysView[str]:
        """
        Gets the keys (tab titles) for the fetched tabs on this GSheet.

        Returns:
            KeysView[str]: The tab titles on this GSheet.

        """
        return self.tabs.keys()

    def values(self) -> ValuesView[Tab]:
        """
        Gets the values (Tabs) for the fetched tabs on this GSheet.

        Returns:
            ValuesView[Tab]: The Tabs on this GSheet.

        """
        return self.tabs.values()

    def get_tab_index_by_title(self, tab_title: str) -> Optional[int]:
        for t in self._tabs:
            if t.title == tab_title:
                return t.index
        else:
            return None

    def to_csv(
        self,
        root_path: str | Path,
        filename_overrides: Dict[str, str] | None = None,
        **tabs_and_headers: Sequence[Any] | None,
    ) -> None:
        """
        Convenience method for calling :meth:`.to_csv() <autodrive.tab.Tab.to_csv>`
        on some or all of the GSheet's tabs.

        Args:
            root_path (str): The root directory path to save tab files to.
            filename_overrides (Dict[str, str], optional): By default, this method
                will name each file after the corresponding tab's title. To
                override some or all of the resulting filenames, pass a dictionary
                with keys equal to the names of the tab you want to override and
                the values equal to the name of the filename you want. Defaults to
                None, for all tabs being treated with default behavior.
            tabs_and_headers: (Sequence[Any], optional): If you want to only
                output the data for some of the tabs, you can pass the names of
                the desired tabs as kwargs. If you wish, you can also pass a
                header row for those tabs, which will be inserted as the first
                row of the file. If you don't want to pass a header row, simply
                pass ``tabname=None`` for that tab.

        Raises:
            ValueError: If root_path does not lead to a directory.
        """
        root_path = Path(root_path)
        if not root_path.is_dir():
            raise ValueError(f"{root_path} is not a directory.")
        overrides = filename_overrides or {}
        tab_info = (
            tabs_and_headers
            if len(tabs_and_headers) > 0
            else {t.title: None for t in self._tabs}
        )
        for tab_name, header in tab_info.items():
            tab = self.tabs[tab_name]
            filename = overrides.get(tab_name, tab.title)
            p = root_path.joinpath(filename).with_suffix(".csv")
            tab.to_csv(p, header)

    def to_json(
        self,
        root_path: str | Path,
        filename_overrides: Dict[str, str] | None = None,
        **tabs_and_headers: Sequence[str] | int,
    ) -> None:
        """
        Convenience method for calling :meth:`.to_json() <autodrive.tab.Tab.to_json>`
        on some or all of the GSheet's tabs.

        Args:
            root_path (str): The root directory path to save tab files to.
            filename_overrides (Dict[str, str], optional): By default, this method
                will name each file after the corresponding tab's title. To
                override some or all of the resulting filenames, pass a dictionary
                with keys equal to the names of the tab you want to override and
                the values equal to the name of the filename you want. Defaults to
                None, for all tabs being treated with default behavior.
            tabs_and_headers: (Sequence[str] | int, optional): If you want to only
                output the data for some of the tabs, you can pass the names of
                the desired tabs as kwargs. You must also indicate what keys
                should be used when creating the jsons for those tabs. For each
                header value, you may either pass a row index to pull for the keys,
                or a list of keys.

        Raises:
            ValueError: If root_path does not lead to a directory.
        """
        root_path = Path(root_path)
        if not root_path.is_dir():
            raise ValueError(f"{root_path} is not a directory.")
        overrides = filename_overrides or {}
        tab_info = (
            tabs_and_headers
            if len(tabs_and_headers) > 0
            else {t.title: 0 for t in self._tabs}
        )
        for tab_name, header in tab_info.items():
            tab = self.tabs[tab_name]
            filename = overrides.get(tab_name, tab.title)
            p = root_path.joinpath(filename).with_suffix(".jsonl")
            tab.to_json(p, header)



from ._conn import Connection
from ._view import GSheetView


View = GSheetView
"""Supertype for Ranges, Tabs, and GSheets."""

Connection = Connection
"""Supertype for SheetsConnections and DriveConnections."""



from __future__ import annotations

from typing import Any, Sequence, Dict

from .connection import SheetsConnection
from ._view import Component
from .formatting.format_rng import (
    RangeCellFormatting,
    RangeGridFormatting,
    RangeTextFormatting,
)
from .interfaces import AuthConfig, FullRange
from .dtypes import EffectiveVal, GoogleValueType


class Range(Component[RangeCellFormatting, RangeGridFormatting, RangeTextFormatting]):
    """
    Provides a connection to the data in a specific range in a Google Sheet Tab.
    """

    def __init__(
        self,
        gsheet_range: FullRange | str,
        gsheet_id: str,
        tab_title: str,
        tab_id: int,
        *,
        auth_config: AuthConfig | None = None,
        sheets_conn: SheetsConnection | None = None,
        autoconnect: bool = True,
    ) -> None:
        """

        Args:
            gsheet_range (FullRange | str): The range (i.e. A5:C10) to associate
                with this Range.
            gsheet_id (str): The id string of the target Google Sheet that the
                Range resides in; can be found in the Google Sheet url.
            tab_title (str): The name of the Tab this Range is within.
            auth_config (AuthConfig, optional): Optional custom AuthConfig,
                defaults to None.
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.
            autoconnect (bool, optional): If you want to instantiate a Range
                without immediately checking your authentication credentials and
                connection to the Google Sheets api, set this to False, defaults
                to True.
        """
        self._tab_title = tab_title
        rng = FullRange(gsheet_range) if isinstance(gsheet_range, str) else gsheet_range
        rng.tab_title = tab_title
        self._rng = rng
        if not self._rng.tab_title:
            self._rng.tab_title = self._tab_title
        super().__init__(
            gsheet_id=gsheet_id,
            tab_id=tab_id,
            gsheet_range=rng,
            grid_formatting=RangeGridFormatting,
            text_formatting=RangeTextFormatting,
            cell_formatting=RangeCellFormatting,
            auth_config=auth_config,
            sheets_conn=sheets_conn,
            autoconnect=autoconnect,
        )

    @property
    def format_grid(self) -> RangeGridFormatting:
        """
        Returns:
            RangeGridFormatting: An object with grid formatting methods.

        """
        return self._format_grid

    @property
    def format_text(self) -> RangeTextFormatting:
        """
        Returns:
            RangeTextFormatting: An object with text formatting methods.

        """
        return self._format_text

    @property
    def format_cell(self) -> RangeCellFormatting:
        """
        Returns:
            RangeCellFormatting: An object with cell formatting methods.

        """
        return self._format_cell

    def get_data(self, value_type: GoogleValueType = EffectiveVal) -> Range:
        """
        Gets the data from the cells of this Range.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            value_type (GoogleValueType, optional): Allows you to toggle the
                type of the values returned by the Google Sheets API. See the
                :mod:`dtypes <autodrive.dtypes>` documentation for more info on
                the different GoogleValueTypes.

        Returns:
          Range: This Range.

        """
        self._values, self._formats = self._get_data(
            self._gsheet_id, str(self._rng), value_type
        )
        return self

    def write_values(self, data: Sequence[Sequence[Any] | Dict[str, Any]]) -> Range:
        """
        Adds a request to write data. Range.commit () to commit the requests.

        Args:
            data (Sequence[Sequence[Any] | Dict[str, Any]]): The data to write.
                Each sequence or dictionary in the passed data is a row, with
                each value in that sub-iterable being a column. Dictionary keys
                will be used as a header row in the written data.

        Returns:
          Range: This Tab.

        """
        self._write_values(data, self._tab_id, self._rng.to_dict())
        return self



from __future__ import annotations

from typing import Any, Dict

from .. import _google_terms as terms
from ..dtypes import HorizontalAlign, UserEnteredFmt, VerticalAlign
from ..interfaces import Format, FullRange


def apply_format(tab_id: int, rng: FullRange, fmt: Format) -> Dict[str, Any]:
    return {
        terms.RPT_CELL: {
            terms.RNG: {terms.TAB_ID: tab_id, **dict(rng)},
            terms.FIELDS: str(fmt),
            terms.CELL: dict(fmt),
        }
    }


def set_text_alignment(
    tab_id: int,
    rng: FullRange,
    halign: HorizontalAlign | None = None,
    valign: VerticalAlign | None = None,
) -> Dict[str, Any]:
    align_dict: Dict[str, str] = {}
    if halign:
        align_dict["horizontalAlignment"] = str(halign)
    if valign:
        align_dict["verticalAlignment"] = str(valign)
    return {
        terms.RPT_CELL: {
            terms.RNG: {terms.TAB_ID: tab_id, **dict(rng)},
            terms.FIELDS: "*",
            terms.CELL: {str(UserEnteredFmt): align_dict},
        }
    }



from __future__ import annotations

import os
from abc import ABC
from typing import List, Literal, Dict, Any, cast, Tuple

from google.auth.transport.requests import Request  # type: ignore
from google.auth.exceptions import RefreshError  # type: ignore
from google.oauth2.credentials import Credentials  # type: ignore
from google_auth_oauthlib.flow import InstalledAppFlow  # type: ignore
from googleapiclient.discovery import Resource, build

from .interfaces import AuthConfig
from . import _google_terms as terms

SCOPES = [
    "https://www.googleapis.com/auth/drive",
    "https://www.googleapis.com/auth/spreadsheets",
]


class Connection(ABC):
    """
    Base class of more specific Connection objects.
    """

    google_obj_types = {
        "folder": "application/vnd.google-apps.folder",
        "sheet": "application/vnd.google-apps.spreadsheet",
        # "doc": "application/vnd.google-apps.document",
        # "slide": "application/vnd.google-apps.presentation",
    }

    def __init__(
        self,
        *,
        api_name: Literal["sheets", "drive"],
        api_version: str,
        auth_config: AuthConfig | None = None,
    ) -> None:
        """

        Args:
          api_name (Literal["sheets", "drive"]): The name of the api to connect to.
          api_version (str): The version of the api to connect to.
          api_scopes (List[str]): A list of the scope strings to connect with
            (will fail if credentials don't grant those scopes).
          auth_config (AuthConfig, optional): Optional custom AuthConfig object,
            defaults to None

        """
        self._auth_config = auth_config or AuthConfig()
        self._core = self._connect(SCOPES, api_name, api_version)

    @property
    def auth(self) -> AuthConfig:
        """
        Returns:
          AuthConfig: The AuthConfig passed to this Connection object, or the
          AuthConfig that was generated automatically upon this Connection's
          instantiation.

        """
        return self._auth_config

    @classmethod
    def _authenticate(
        cls,
        scopes: List[str],
        auth_config: AuthConfig,
    ) -> Credentials:
        """
        Uses configuration in passed AuthConfig to attempt to login to Google
        Drive. The first time it is run it will cause a web page to open up and
        solicit permission to access the Google Drive as specified in the
        credentials. Then it will create a token that it will use going
        forward.

        Note, you can also set environment variables as described in
        Connection.get_creds_from_env () and _authenticate will automatically
        pull those in and will not create a token file.

        Args:
          scopes (List[str]): A list of scopes dictating the limits of the
            authenticated connection.
          auth_config (AuthConfig): An AuthConfig object, either customized or
            automatically generated.

        Returns:
          Credentials: A prepped Credentials object, ready to be slotted into more
          specific connections.

        Raises:
          FileNotFoundError: If no credentials file is found and no creds were
          passed as environment variables or as a dictionary within AuthConfig.

        """
        creds = cls.get_creds_from_env()
        # The token file stores the user's access and refresh tokens, and
        # is created automatically when the authorization flow completes for the
        # first time.
        if not creds and auth_config.token_filepath.exists():
            creds = Credentials.from_authorized_user_file(  # type: ignore
                auth_config.token_filepath, scopes
            )
        if creds and creds.expired and creds.refresh_token:  # type: ignore
            try:
                creds.refresh(Request())  # type: ignore
            except RefreshError:
                creds = None
        # If there are no (valid) credentials available, prompt user to log in.
        if not creds or not creds.valid:
            if auth_config.secrets_config:
                flow = InstalledAppFlow.from_client_config(  # type: ignore
                    auth_config.secrets_config, scopes
                )
            elif auth_config.creds_filepath.exists():
                flow = InstalledAppFlow.from_client_secrets_file(  # type: ignore
                    str(auth_config.creds_filepath), scopes
                )
            else:
                raise FileNotFoundError(
                    f"Credentials file {auth_config.creds_filepath} could not be "
                    "found."
                )
            creds = flow.run_local_server(port=0)  # type: ignore
            # Save the credentials for the next run
            with open(auth_config.token_filepath, "w") as token:
                token.write(creds.to_json())  # type: ignore
        return creds

    def _connect(
        self, scopes: List[str], api: Literal["drive", "sheets"], version: str
    ) -> Resource:
        """
        Generates a connection to the specified Google api.

        Args:
          scopes (List[str]): A list of scopes dictating the limits of the
            authenticated connection
          api (Literal["sheets", "drive"]): The name of the api to connect to.
          version (str): The version of the api to connect to.

        Returns:
          Resource: A Google api client Resource, which is a very general (and
          poorly type annotated/highly dynamic) object that facilitates connections
          to the appropriate Google api.

        """
        creds = self._authenticate(
            scopes,
            self._auth_config,
        )
        return build(api, version, credentials=creds)

    @staticmethod
    def get_creds_from_env() -> Credentials | None:
        """
        Checks if necessary credential and token information has been specified
        in environment variables and uses that information to generate a Credentials
        object. Can be used to bypass needing credentials files saved locally.

        Returns:
          Credentials, optional: A prepped Credentials object, ready to be slotted
          into more specific connections, if appropriate environment variables were
          found, otherwise None.

        """
        token = os.getenv("AUTODRIVE_TOKEN")
        refresh_token = os.getenv("AUTODRIVE_REFR_TOKEN")
        client_id = os.getenv("AUTODRIVE_CLIENT_ID")
        client_secret = os.getenv("AUTODRIVE_CLIENT_SECRET")
        if token and refresh_token and client_id and client_secret:
            return Credentials(
                token=token,
                refresh_token=refresh_token,
                token_uri="https://oauth2.googleapis.com/token",
                client_id=client_id,
                client_secret=client_secret,
                scopes=["https://www.googleapis.com/auth/drive"],
            )
        else:
            return None

    @classmethod
    def _merge_dicts(
        cls, dict1: Dict[str, Any], dict2: Dict[str, Any]
    ) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        keys = {*dict1.keys(), *dict2.keys()}
        for k in keys:
            val1 = dict1.get(k)
            val2 = dict2.get(k)
            final_val: Any = val1
            if isinstance(val1, dict) and isinstance(val2, dict):
                final_val = cls._merge_dicts(
                    cast(Dict[str, Any], val1), cast(Dict[str, Any], val2)
                )
            elif val2 is not None:
                final_val = val2
            result[k] = final_val
        return result

    @staticmethod
    def _create_range_tuple_key(input: Dict[str, Any]) -> Tuple[Tuple[str, Any], ...]:
        key_order = (
            terms.TAB_ID,
            terms.STARTROW,
            terms.ENDROW,
            terms.STARTCOL,
            terms.ENDCOL,
            terms.STARTIDX,
            terms.ENDIDX,
        )
        result: List[Tuple[str, Any]] = []
        for key in key_order:
            if key in input.keys():
                result.append((key, input[key]))
        return tuple(result)

    @classmethod
    def _preprocess_requests(
        cls, requests: List[Dict[str, Any]]
    ) -> Dict[str, List[Dict[str, Any]]]:
        ranged_requests: Dict[
            str, Dict[Tuple[Tuple[str, Any], ...], Dict[str, Any]]
        ] = {}
        result: List[Dict[str, Any]] = []
        for request in requests:
            range_key = None
            for request_type in request.keys():
                if terms.RNG in request[request_type].keys():
                    range_key = cls._create_range_tuple_key(
                        request[request_type][terms.RNG]
                    )
                if range_key:
                    if request_type not in ranged_requests.keys():
                        ranged_requests[request_type] = {}
                    if range_key in ranged_requests[request_type].keys():
                        existing_dict = ranged_requests[request_type][range_key]
                    else:
                        existing_dict = {}
                    ranged_requests[request_type][range_key] = cls._merge_dicts(
                        existing_dict, request
                    )
                else:
                    result.append(request)
        for range_dict in ranged_requests.values():
            result += list(range_dict.values())
        return {"requests": result}



from . import connection, dtypes, interfaces, formatting
from .drive import Drive, Folder
from .dtypes import (
    Boolean,
    EffectiveVal,
    FormattedVal,
    Formula,
    Number,
    String,
    UserEnteredVal,
)
from .gsheet import GSheet
from .interfaces import (
    AccountingFormat,
    AuthConfig,
    Color,
    NumericFormat,
    HalfRange,
    NumberFormat,
    TextFormat,
    FullRange,
)
from .range import Range
from .tab import Tab
from .connection import FileUpload

__all__ = [
    "connection",
    "formatting",
    "Folder",
    "Drive",
    "dtypes",
    "String",
    "Formula",
    "Number",
    "Boolean",
    "UserEnteredVal",
    "EffectiveVal",
    "FormattedVal",
    "GSheet",
    "interfaces",
    "AuthConfig",
    "HalfRange",
    "FullRange",
    "Color",
    "TextFormat",
    "NumericFormat",
    "NumberFormat",
    "AccountingFormat",
    "Range",
    "Tab",
    "FileUpload",
]



from __future__ import annotations

import string
from abc import ABC
from typing import Any, Dict, Generic, List, Tuple, Type, TypeVar, Sequence
from pathlib import Path
import csv
import jsonlines  # type: ignore

from . import _google_terms as terms
from .connection import SheetsConnection
from .dtypes import (
    GOOGLE_DTYPES,
    TYPE_MAP,
    EffectiveFmt,
    EffectiveVal,
    Formula,
    String,
    GoogleValueType,
    UserEnteredVal,
)
from .interfaces import AuthConfig, FullRange
from ._core import GoogleDtype

T = TypeVar("T", bound="GSheetView")
FC = TypeVar("FC", bound="CellFormatting")
FG = TypeVar("FG", bound="GridFormatting")
FT = TypeVar("FT", bound="TextFormatting")


class NoConnectionError(Exception):
    """
    Error thrown when attempting to connect via a SheetsConnection that does not
    exist.
    """

    def __init__(self, vtype: Type[GSheetView], *args: object) -> None:
        msg = f"No SheetsConnection has been established for this {vtype}."
        super().__init__(msg, *args)


class OutputError(Exception):
    """
    Error thrown when something goes wrong when attempting to write values to a
    file.
    """


class Formatting:
    """
    Formatting objects contain methods for adding various update format requests
    to their parent Component.
    """

    def __init__(self, parent: Component[Any, Any, Any]):
        """

        Args:
            parent (Component[Any, Any, Any]): A Component object.

        """
        self._parent = parent

    def add_request(self, request: Dict[str, Any]) -> None:
        """
        Adds the passed request to the Formatting object's parent component.

        Args:
          request (Dict[str, Any]): An api-ready request.

        """
        self._parent.requests.append(request)

    def ensure_full_range(self, rng: FullRange | str | None = None) -> FullRange:
        """
        Convenience method for ensuring that all requests generated by this
        Formatting object have a FullRange attached to them, if one isn't manually
        supplied.

        Args:
            rng (FullRange | str, optional): A manually generated FullRange, defaults to
                None.

        Returns:
            FullRange: The passed FullRange, or a range generated from the
            Formatting object's parent Component.

        """
        return self._parent.ensure_full_range(self._parent.range, rng)


class CellFormatting(Formatting):
    """
    Contains methods for generating format requests that update cell properties.
    """

    ...


class GridFormatting(Formatting):
    """
    Contains methods for generating format requests that update grid properties.
    """

    ...


class TextFormatting(Formatting):
    """
    Contains methods for generating format requests that update the properties of
    text in one or more cells.
    """

    ...


class GSheetView(ABC):
    """
    Abstract base class for the different ways of viewing a Google Sheet.
    """

    def __init__(
        self,
        gsheet_id: str,
        *,
        auth_config: AuthConfig | None = None,
        sheets_conn: SheetsConnection | None = None,
        autoconnect: bool = True,
    ) -> None:
        """

        Args:
            gsheet_id (str): The id string of the target Google Sheet that the View
                is attached to.
            auth_config (AuthConfig, optional): Optional custom AuthConfig, defaults
                to None.
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.
            autoconnect (bool, optional): If you want to instantiate a View without
                immediately checking your authentication credentials and connection
                to the Google Sheets api, set this to False, defaults to True.
        """
        super().__init__()
        if not sheets_conn and autoconnect:
            sheets_conn = SheetsConnection(auth_config=auth_config)
        self._conn = sheets_conn
        self._auth = auth_config
        self._gsheet_id = gsheet_id
        self._requests: List[Dict[str, Any]] = []

    @property
    def requests(self) -> List[Dict[str, Any]]:
        """
        Returns:
            List[Dict[str, Any]]: The list of current (uncommitted) requests.

        """
        return self._requests

    @property
    def conn(self) -> SheetsConnection:
        """
        Returns:
            SheetsConnection: The view's SheetsConnection.

        Raises:
            NoConnectionError: If the view's connection is null.

        """
        if not self._conn:
            raise NoConnectionError(type(self))
        return self._conn

    @property
    def auth(self) -> AuthConfig:
        """
        Returns:
            AuthConfig: The view's AuthConfig.

        Raises:
            NoConnectionError: If the view's auth config is null.

        """
        if not self._auth:
            raise NoConnectionError(type(self))
        return self._auth

    @property
    def gsheet_id(self) -> str:
        """
        Returns:
            str: The id of the Google Sheet this view is connected to.

        """
        return self._gsheet_id

    def commit(self) -> Dict[str, Any]:
        """
        Commits the amassed requests on this view, sending them to the Sheets api
        as a batch update request.

        Returns:
            Dict[str, Any]: The response from the api.

        Raises:
            NoConnectionError: If the view's SheetsConnection is null.

        """
        if not self._conn:
            raise NoConnectionError(type(self))
        results = self._conn.execute_requests(self._gsheet_id, self._requests)
        self._requests = []
        return results

    def ensure_full_range(
        self, backup: FullRange, rng: FullRange | str | None = None
    ) -> FullRange:
        """
        Convenience method for ensuring that a range argument that could be None
        or a string is always a FullRange.

        Args:
          rng (FullRange | str, optional): A manually generated FullRange or
            valid FullRange string, defaults to None.

        Returns:
            FullRange: The passed FullRange, or the backup FullRange.

        """
        if isinstance(rng, str):
            rng = FullRange(rng)
        return rng if rng else backup

    @staticmethod
    def _parse_properties(
        properties: Dict[str, Any]
    ) -> Tuple[str, List[Dict[str, Any]]]:
        sheet_title = properties[terms.FILE_PROPS][terms.FILE_NAME]
        sheet_props = [sheet[terms.TAB_PROPS] for sheet in properties[terms.TABS_PROP]]
        return sheet_title, sheet_props

    @classmethod
    def _parse_row_data(
        cls,
        row_data: List[Dict[str, List[Dict[str, Any]]]],
        value_type: GoogleValueType = EffectiveVal,
    ) -> Tuple[List[List[Any]], List[List[Dict[str, Any]]]]:
        """
        Parses the dictionary returned by SheetsConnection.get_data and extracts
        only the relevant data.

        Args:
            row_data (List[Dict[str, List[Dict[str, Any]]]]): The raw data to parse.
            value_type (GoogleValueType, optional): The value representation to
                extract from the raw data, defaults to EffectiveVal

        Returns:
            Tuple[List[List[Any]], List[List[Dict[str, Any]]]]: A tuple containing
            a list of extracted data and another list of extracted formatting
            information.

        """
        values: List[List[Any]] = []
        formats: List[List[Dict[str, Any]]] = []
        for row in row_data:
            value_list: List[Any] = []
            fmt_list: List[Dict[str, Any]] = []
            for cell in row.get(terms.VALUES, []):
                raw_value = cell.get(str(value_type))
                fmt = cell.get(str(EffectiveFmt), {})
                value = raw_value
                if value_type in (UserEnteredVal, EffectiveVal):
                    if raw_value:
                        for dtype in GOOGLE_DTYPES:
                            value = raw_value.get(str(dtype))
                            if value:
                                if value_type == UserEnteredVal:  # type: ignore
                                    value = dtype.parse(value)
                                break
                value_list.append(value)
                fmt_list.append(fmt)
            values.append(value_list)
            formats.append(fmt_list)
        return values, formats

    def _get_data(
        self,
        gsheet_id: str,
        rng_str: str,
        value_type: GoogleValueType = EffectiveVal,
    ) -> Tuple[List[List[Any]], List[List[Dict[str, Any]]]]:
        """
        Fetches data from the view's SheetsConnection for the specified Google
        Sheet and parses it.

        Args:
            gsheet_id (str): The Google Sheet to fetch data from.
            rng_str (str: str): The range within the Google Sheet to fetch data from.
            value_type (GoogleValueType, optional): The value representation to
                extract from the raw data, defaults to EffectiveVal

        Returns:
            Tuple[List[List[Any]], List[List[Dict[str, Any]]]]: A tuple containing
            a list of data values and another list of formatting information.

        """
        raw = self.conn.get_data(gsheet_id, [rng_str])
        row_data = raw[terms.TABS_PROP][0][terms.DATA][0].get(terms.ROWDATA, [])
        return self._parse_row_data(row_data, value_type=value_type)

    def _write_values(
        self: T,
        data: Sequence[Sequence[Any] | Dict[str, Any]],
        tab_id: int,
        rng_dict: Dict[str, int] | None = None,
    ) -> T:
        """
        Generates an update or append cells request for writing values to the
        target sheet.

        Args:
            data (Sequence[Sequence[Any] | Dict[str, Any]]): The data to write.
            rng_dict (Dict[str, int]): The range properties to write to. Defaults
                to None, in which case the values will be appended after the last
                populated row of the sheet.

        Returns:
            T: This View object.

        """
        write_values: List[List[Dict[str, Any]]] = []
        for row in data:
            if isinstance(row, dict):
                if len(write_values) == 0:
                    write_values.append(
                        [self._gen_cell_write_value(k) for k in row.keys()]
                    )
                std_row = [val for val in row.values()]
            else:
                std_row = [val for val in row]
            write_values.append([self._gen_cell_write_value(val) for val in std_row])
        if rng_dict is not None:
            target = {terms.RNG: {terms.TAB_ID: tab_id, **rng_dict}}
            req_type = "updateCells"
        else:
            target = {terms.TAB_ID: tab_id}
            req_type = "appendCells"
        request = {
            req_type: {
                terms.FIELDS: "*",
                terms.ROWS: [{terms.VALUES: values} for values in write_values],
                **target,
            }
        }
        self._requests.append(request)
        return self

    @staticmethod
    def _gen_cell_write_value(python_val: Any) -> Dict[str, Any]:
        """
        Converts a python value to its corresponding Google Dtype and wraps it in
        the key-value payload expected by the Sheets api.

        Args:
            python_val (Any): Any value. Only numeric values, booleans, and
                "formulas" (strings that start with =) will be converted. All
                other python data will be converted to its string representation.

        Returns:
            Dict[str, Any]: A dictionary containing the details the Sheet's api
            expects when writing data.

        """
        dtype: GoogleDtype
        type_ = type(python_val)
        if (
            isinstance(python_val, str)
            and len(python_val) >= 2
            and python_val[0] == "="
        ):
            dtype = Formula
        else:
            dtype = TYPE_MAP.get(type_, String)
        return {UserEnteredVal.value_key: {dtype.type_key: python_val}}

    # TODO: Delete this?
    @staticmethod
    def gen_alpha_keys(num: int) -> List[str]:
        """
        Generates a list of characters from the Latin alphabet a la gsheet/excel
        headers.

        Args:
            num (int): The desired length of the list.

        Returns:
            List[str]: A list containing as many letters and letter combos as
            desired. Can be used to generate sets up to 676 in length.

        """
        a = string.ascii_uppercase
        result: List[str] = list()
        x = num // 26
        for i in range(x + 1):
            root = a[i - 1] if i > 0 else ""
            keys = [root + a[j] for j in range(26)]
            for k in keys:
                result.append(k) if len(result) < num else None
        return result


class Component(GSheetView, Generic[FC, FG, FT]):
    """
    Base class for Tab and Range.
    """

    def __init__(
        self,
        *,
        gsheet_range: FullRange,
        gsheet_id: str,
        tab_id: int,
        grid_formatting: Type[FG],
        text_formatting: Type[FT],
        cell_formatting: Type[FC],
        auth_config: AuthConfig | None = None,
        sheets_conn: SheetsConnection | None = None,
        autoconnect: bool = True,
    ) -> None:
        super().__init__(
            gsheet_id,
            auth_config=auth_config,
            sheets_conn=sheets_conn,
            autoconnect=autoconnect,
        )
        self._rng = gsheet_range
        self._tab_id = tab_id
        self._values: List[List[Any]] = []
        self._formats: List[List[Dict[str, Any]]] = []
        self._format_grid = grid_formatting(self)
        self._format_text = text_formatting(self)
        self._format_cell = cell_formatting(self)

    @property
    def tab_id(self) -> int:
        """
        Returns:
            int: The id of the linked tab.

        """
        return self._tab_id

    @property
    def range_str(self) -> str:
        """
        str: The string representation of the range specified by this Component.

        """
        return str(self._rng)

    @property
    def range(self) -> FullRange:
        """
        Returns:
            FullRange: The FullRange representation of the range specified by this
            Component.

        """
        return self._rng

    @property
    def format_grid(self) -> FG:
        """
        Returns:
            FG: The GridFormatting subclass associated with this Component type.

        """
        return self._format_grid

    @property
    def format_text(self) -> FT:
        """
        Returns:
            FT: The TextFormatting subclass associated with this Component type.

        """
        return self._format_text

    @property
    def format_cell(self) -> FC:
        """
        Returns:
            FC: The CellFormatting subclass associated with this Component type.

        """
        return self._format_cell

    @property
    def values(self) -> List[List[Any]]:
        """
        Returns:
            List[List[Any]]: The fetched data values in this Component's cells.

        """
        return self._values

    @values.setter
    def values(self, new_values: List[List[Any]]) -> None:
        """
        Args:
            new_values (List[List[Any]]): The values to overwrite this Component's
                current values with.

        """
        self._values = new_values

    @property
    def formats(self) -> List[List[Dict[str, Any]]]:
        """
        Returns:
            List[List[Dict[str, Any]]]: The fetched formatting properties of this
            Component's cells.

        """
        return self._formats

    @formats.setter
    def formats(self, new_formats: List[List[Dict[str, Any]]]) -> None:
        """
        Args:
            new_formats (List[List[Dict[str, Any]]]): The formats to overwrite this
                Component's current formats with.

        """
        self._formats = new_formats

    @property
    def data_shape(self) -> Tuple[int, int]:
        """
        Returns:
            Tuple[int, int]: The row length and column width of this Component's
            data.

        """
        width = len(self._values[0]) if self._values else 0
        return len(self._values), width

    def to_csv(self, p: str | Path, header: Sequence[Any] | None = None) -> None:
        """
        Saves values to a csv file.

        Args:
            p (str | Path): The path-like for the file to save the data to.
            header (Sequence[Any], optional): A header row. If supplied, must be
                the same number of columns as the data values. Defaults to None.
        """
        if header:
            self._verify_header_len(header)
            values = [list(header), *self._values]
        else:
            values = self._values
        with open(p, "w") as file:
            writer = csv.writer(file)
            writer.writerows(values)

    def to_json(self, p: str | Path, header: Sequence[str] | int = 0) -> None:
        """
        Saves values to a json file, with one json per line.

        Args:
            p (str | Path): The path-like for the file to save the data to.
            header (Sequence[Any] | int): A header row or the index of a row in
                the values data to use as the header row. The header will be used
                as the keys for the json-formatted dictionaries. Defaults to 0,
                using the first row as the header.
        """
        if isinstance(header, Sequence):
            self._verify_header_len(header)
            values = self._values
        else:
            values = self.values
            header = values.pop(header)
        values_dicts = [dict(zip(header, row)) for row in values]
        with jsonlines.open(p, "w") as writer:  # type: ignore
            writer.write_all(values_dicts)  # type: ignore

    def _verify_header_len(self, header: Sequence[Any]) -> bool:
        """
        Ensures the passed header is the appropriate length given the Component's
        data_shape.

        Args:
            header (Sequence[Any]): A sequence of values to check.

        Raises:
            OutputError: If the length of header is less than the width of the
                Component's data.

        Returns:
            bool: True if the header passed verification.
        """
        if len(header) < self.data_shape[1]:
            raise OutputError(
                f"Supplied header length {len(header)} is insufficient for data "
                f"width {self.data_shape[1]}."
            )
        else:
            return True



from __future__ import annotations

from typing import List, Optional, Dict
from pathlib import Path

from .connection import AuthConfig, DriveConnection, SheetsConnection, FileUpload
from .gsheet import GSheet


class Folder:
    def __init__(
        self,
        folder_id: str,
        name: str,
        *,
        parents: List[str] | None = None,
        auth_config: AuthConfig | None = None,
        drive_conn: DriveConnection | None = None,
        sheets_conn: SheetsConnection | None = None,
    ) -> None:
        """
        Stores all the necessary properties of a Google Drive Folder and provides
        methods for interacting with its subfolders and files.

        Args:
            folder_id (str): The id string of the folder.
            name (str): The name of the folder.
            parents (List[str], optional): A list of the Folder's parents, usually
                just its parent folder or parent shared drive, if any, defaults to
                None, signifying that the folder is at the root of an unshared
                drive.
            auth_config (AuthConfig, optional): Optional custom AuthConfig object,
                defaults to None.
            drive_conn (DriveConnection, optional): Optional manually created
                DriveConnection, defaults to None
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.

        """
        self._id = folder_id
        self._name = name
        self._parents = parents or []
        self._conn = drive_conn or DriveConnection(auth_config=auth_config)
        self._sheets_conn = sheets_conn or SheetsConnection(auth_config=self._conn.auth)

    @property
    def id(self) -> str:
        """
        Returns:
            str: The id string of the Folder.

        """
        return self._id

    @property
    def name(self) -> str:
        """
        Returns:
            str: The name of the Folder.

        """
        return self._name

    @property
    def parent_ids(self) -> List[str]:
        """
        Returns:
            List[str]: The list of the Folder's parents, usually just its parent
            folder or parent shared drive, if any.

        """
        return self._parents

    def create_folder(self, folder_name: str) -> Folder:
        """
        Generates a new Folder with this Folder as its parent.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            folder_name (str): The name of the Folder to create.

        Returns:
            Folder: The newly created Folder.

        """
        id = self._conn.create_object(folder_name, "folder", self._id)
        return Folder(
            id,
            folder_name,
            parents=[self._id],
            drive_conn=self._conn,
        )

    def create_gsheet(self, gsheet_name: str) -> GSheet:
        """
        Creates a new Google Sheet with this Folder as its parent.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            gsheet_name (str): The desired name of the new Google Sheet.

        Returns:
            GSheet: The newly created GSheet.

        """
        new_id = self._conn.create_object(gsheet_name, "sheet", self._id)
        return GSheet(new_id, gsheet_name, sheets_conn=self._sheets_conn)

    def upload_files(self, *filepaths: Path | str | FileUpload) -> Dict[str, str]:
        """
        Uploads files to the folder.

        Args:
            *filepaths (Path | str | FileUpload): An arbitrary number of Path
                objects, path strings, or FileUpload objects (if you want to
                convert the file to a Google Drive format (e.g. Docs, Sheets,
                etc).

        Returns:
            Dict[str, str]: The names of the uploaded files and their new ids in
                Google Drive.
        """
        fps: List[FileUpload] = []
        for fp in filepaths:
            if isinstance(fp, FileUpload):
                fp.folder = self.id
            else:
                fp = FileUpload(fp, self.id)
        return self._conn.upload_files(*fps)


class Drive:
    """
    A one-stop shop for finding and managing Google Drive objects, including
    folders and Google Sheets.
    """

    def __init__(
        self,
        *,
        auth_config: AuthConfig | None = None,
        drive_conn: DriveConnection | None = None,
        sheets_conn: SheetsConnection | None = None,
    ) -> None:
        """
        .. note::

            All the methods on this class will cause a request to be posted to the
            relevant Google API immediately.

        Args:
            auth_config (AuthConfig, optional): Optional custom AuthConfig object,
                defaults to None.
            drive_conn (DriveConnection, optional): Optional manually created
                DriveConnection, defaults to None.
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.

        """
        self._conn = drive_conn or DriveConnection(auth_config=auth_config)
        self._sheets_conn = sheets_conn or SheetsConnection(auth_config=self._conn.auth)

    def create_folder(
        self, folder_name: str, parent: str | Folder | None = None
    ) -> Folder:
        """
        Creates a folder in the connected Google Drive.

        Args:
            folder_name (str): The desired name of the new folder
            parent (str | Folder, optional): The parent folder/id to create this
                folder within, or the id of the shared drive to create within,
                defaults to None.

        Returns:
            Folder: The newly created Folder.

        """
        parent_id = self._ensure_parent_id(parent)
        new_id = self._conn.create_object(folder_name, "folder", parent_id)
        return Folder(
            folder_id=new_id,
            name=folder_name,
            parents=[parent_id] if parent_id else None,
            drive_conn=self._conn,
        )

    def create_gsheet(
        self, gsheet_name: str, parent: str | Folder | None = None
    ) -> GSheet:
        """
        Creates a Google Sheet in the connected Google Drive.

        Args:
            gsheet_name (str): The desired name of the new Google Sheet.
            parent (str | Folder, optional): The parent folder/id to create this
                gsheet within, or the id of the shared drive to create within,
                defaults to None.

        Returns:
            GSheet: The newly created GSheet.

        """
        parent_id = self._ensure_parent_id(parent)
        new_id = self._conn.create_object(gsheet_name, "sheet", parent_id)
        return GSheet(new_id, gsheet_name, sheets_conn=self._sheets_conn)

    def find_gsheet(self, gsheet_name: str) -> List[GSheet]:
        """
        Search for Google Sheets that match the passed name. GSheets found in this
        way will need to have their properties gathered with a GSheet.fetch () call.
        To save round trips this method only collects the basic details of the Google
        Sheets it matches. These properties cannot be gathered via the Google Drive
        api, unfortunately.

        Args:
            gsheet_name (str): The name (or part of the name) to search for.

        Returns:
            List[GSheet]: A list of GSheet objects with basic details (id, name)
            that matched the name parameter.

        """
        result = self._conn.find_object(gsheet_name, "sheet")
        return [
            GSheet(r["id"], r["name"], sheets_conn=self._sheets_conn) for r in result
        ]

    def find_folder(
        self, folder_name: str, shared_drive_id: str | None = None
    ) -> List[Folder]:
        """
        Search for folders that match the passed name.

        Args:
            folder_name (str): The name (or part of the name) to search for.
            shared_drive_id (str, optional): The shared drive id to search within,
                defaults to None

        Returns:
            List[Folder]: A list of Folder object that matched the name parameter.

        """
        result = self._conn.find_object(folder_name, "folder", shared_drive_id)
        return [
            Folder(r["id"], r["name"], parents=r["parents"], drive_conn=self._conn)
            for r in result
        ]

    def upload_files(self, *filepaths: Path | str | FileUpload) -> Dict[str, str]:
        """
        Uploads files to the root drive or to a folder.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            *filepaths (Path | str | FileUpload): An arbitrary number of Path
                objects, path strings, or FileUpload objects (for more complex
                cases).

        Returns:
            Dict[str, str]: The names of the uploaded files and their new ids in
                Google Drive.
        """
        return self._conn.upload_files(*filepaths)

    @staticmethod
    def _ensure_parent_id(parent: str | Folder | None = None) -> Optional[str]:
        """
        Simple convenience method for standardizing parent inputs to be ids and not
        Folders.

        Args:
            parent (str | Folder, optional): Parent id or Folder to extract parent_id
                from, defaults to None

        Returns:
            Optional[str]: The extracted parent_id or None.

        """
        return parent.id if isinstance(parent, Folder) else parent



from __future__ import annotations

from typing import Any, Dict, overload

from .. import _google_terms as terms
from ..interfaces import HalfRange


def auto_column_width(tab_id: int, rng: HalfRange) -> Dict[str, Any]:
    return {
        "autoResizeDimensions": {
            terms.DIMS: {
                terms.TAB_ID: tab_id,
                **dict(rng),
                terms.DIM: terms.COLDIM,
            }
        }
    }


def append_rows(tab_id: int, num_rows: int) -> Dict[str, Any]:
    return {
        "appendDimension": {
            terms.TAB_ID: tab_id,
            terms.DIM: terms.ROWDIM,
            "length": num_rows,
        }
    }


def insert_rows(tab_id: int, num_rows: int, at_row: int) -> Dict[str, Any]:
    return {
        "insertDimension": {
            terms.RNG: {
                terms.TAB_ID: tab_id,
                **dict(HalfRange(at_row, at_row + num_rows, base0_idxs=True)),
                terms.DIM: terms.ROWDIM,
            },
            "inheritFromBefore": False,
        }
    }


def delete_rows(tab_id: int, rng: HalfRange) -> Dict[str, Any]:
    return {
        "deleteDimension": {
            terms.RNG: {
                terms.TAB_ID: tab_id,
                **dict(rng),
                terms.DIM: terms.ROWDIM,
            }
        }
    }


@overload
def freeze(tab_id: int, *, rows: int, columns: int | None = None) -> Dict[str, Any]:
    ...


@overload
def freeze(tab_id: int, *, columns: int, rows: int | None = None) -> Dict[str, Any]:
    ...


def freeze(
    tab_id: int, *, rows: int | None = None, columns: int | None = None
) -> Dict[str, Any]:
    grid_prop = {}
    if not rows and not columns:
        raise ValueError("One of rows or columns must not be None.")
    if rows:
        grid_prop["frozenRowCount"] = rows
    if columns:
        grid_prop["frozenColumnCount"] = columns
    return {
        "updateSheetProperties": {
            terms.TAB_PROPS: {terms.TAB_ID: tab_id, terms.GRID_PROPS: grid_prop},
            terms.FIELDS: f"{terms.GRID_PROPS}(frozenRowCount, frozenColumnCount)",
        }
    }


def append_columns(tab_id: int, num_cols: int) -> Dict[str, Any]:
    return {
        "appendDimension": {
            terms.TAB_ID: tab_id,
            terms.DIM: terms.COLDIM,
            "length": num_cols,
        }
    }


def insert_columns(tab_id: int, num_cols: int, at_col: int) -> Dict[str, Any]:
    return {
        "insertDimension": {
            terms.RNG: {
                terms.TAB_ID: tab_id,
                **dict(HalfRange(at_col, at_col + num_cols, base0_idxs=True)),
                terms.DIM: terms.COLDIM,
            },
            "inheritFromBefore": False,
        }
    }


def delete_columns(tab_id: int, rng: HalfRange) -> Dict[str, Any]:
    return {
        "deleteDimension": {
            terms.RNG: {
                terms.TAB_ID: tab_id,
                **dict(rng),
                terms.DIM: terms.COLDIM,
            }
        }
    }



from __future__ import annotations

from typing import List

from .._view import CellFormatting, GridFormatting, TextFormatting
from ..interfaces import Color, Format, BorderFormat
from . import _cell as cell, _grid as grid, _text as text
from ..dtypes import (
    BorderSide,
    BorderStyle,
    BorderSides,
    HorizontalAlign,
    VerticalAlign,
)


class RangeCellFormatting(CellFormatting):
    """
    Contains request generation methods related to formatting a Range's grid
    (width and height, etc).
    """

    def add_alternating_row_background(self, colors: Color) -> RangeCellFormatting:
        """
        Queues a request to add an alternating row background of the indicated
        color to a Range's cells.

        Args:
            colors (Color): The desired Color to apply to every other row.

        Returns:
            RangeCellFormatting: This formatting object, so further requests can
            be queued if desired.

        """
        self.add_request(
            cell.add_alternating_row_background(
                self._parent.tab_id, self._parent.range, colors
            )
        )
        return self

    def set_background_color(self, color: Color) -> RangeCellFormatting:
        """
        Queues a request to set the background of the Range's cells to the
        indicated color.

        Args:
            color (Color): The desired Color to set the background to.

        Returns:
            RangeCellFormatting: This formatting object, so further requests can
            be queued if desired.
        """
        self.add_request(
            cell.set_background_color(self._parent.tab_id, self._parent.range, color)
        )
        return self

    def set_border_format(
        self,
        *sides: BorderSide,
        style: BorderStyle | None = None,
        color: Color | None = None
    ) -> RangeCellFormatting:
        """
        Queues a request to set the border properties of the Range's cells.

        Args:
            *sides: (BorderSide): One or more BorderSide objects, indicating
                which side(s) of the cells you want to apply the border
                properties to. If no sides are provided, set_border_format will
                apply border properties to all sides.
            style (BorderStyle, optional): The style to apply to all the
                indicated sides. Defaults to None, for the default border style.
            color (Color, optional): The color to set the border(s) to. Defaults
                to None, for black.

        Returns:
            RangeCellFormatting: This formatting object, so further requests can
            be queued if desired.
        """
        fmts: List[BorderFormat] = []
        if len(sides) == 0:
            sides = BorderSides
        for side in sides:
            fmts.append(BorderFormat(side, color, style))
        self.add_request(
            cell.set_border_format(self._parent.tab_id, self._parent.range, *fmts)
        )
        return self


class RangeGridFormatting(GridFormatting):
    """
    Contains request generation methods relating to formatting a Range's
    cells (like adding borders and backgrounds and such).
    """

    def auto_column_width(self) -> RangeGridFormatting:
        """
        Queues a request to set the column width of the Range's columns equal to
        the width of the values in the cells.

        Returns:
            RangeGridFormatting: This formatting object, so further requests can
            be queued if desired.

        """
        self.add_request(
            grid.auto_column_width(self._parent.tab_id, self._parent.range.col_range)
        )
        return self


class RangeTextFormatting(TextFormatting):
    """
    Contains request generation methods methods relating to formatting this
    Range's text (the text format of any cells, even those containing non-text
    values like integers or null values).
    """

    def apply_format(self, format: Format) -> RangeTextFormatting:
        """
        Queues a request to set the text/number format of the Range's cells.

        Args:
          format (Format): A format instance, such as TextFormat or NumberFormat.

        Returns:
            RangeTextFormatting: This formatting object, so further requests can
            be queued if desired.

        """
        self.add_request(
            text.apply_format(self._parent.tab_id, self._parent.range, format)
        )
        return self

    def set_alignment(
        self, *aligns: HorizontalAlign | VerticalAlign
    ) -> RangeTextFormatting:
        """
        Queues a request to set the horizontal and/or vertical text alignment of
        the Range's cells.

        Args:
            aligns (HorizontalAlign | VerticalAlign): The desired horizontal
                and/or vertical alignment properties. Note that if you specify
                a HorizontalAlign more than once, or a VerticalAlign more than
                once, only the last of each will be used.

        Returns:
            RangeTextFormatting: This formatting object, so further requests can
            be queued if desired.
        """
        halign = None
        valign = None
        for align in aligns:
            if isinstance(align, HorizontalAlign):
                halign = align
            else:
                valign = align
        self.add_request(
            text.set_text_alignment(
                self._parent.tab_id, self._parent.range, halign, valign
            )
        )
        return self



# All the keys and such that Google uses in its Drive and Sheets APIS that are
# relevant to Autodrive are listed here. If they ever change, it will be easy to
# change them here and alter all modules to reflect the update.

# General Google API query values:
FIELDS = "fields"
# Google Drive object property names:
ID = "id"
NAME = "name"
PARENTS = "parents"
# Google Sheet property names:
FILE_NAME = "title"
FILE_PROPS = "properties"
TABS_PROP = "sheets"
# Tab property names:
TAB_PROPS = "properties"
TAB_IDX = "index"
TAB_ID = "sheetId"
TAB_NAME = "title"
GRID_PROPS = "gridProperties"
COL_CT = "columnCount"
ROW_CT = "rowCount"
# Data property names:
DATA = "data"
ROWDATA = "rowData"
VALUES = "values"
# Requests:
ADDTAB = "addSheet"
RPT_CELL = "repeatCell"
# Update cell property names:
ROWS = "rows"
RNG = "range"
CELL = "cell"
RNG1D = "range"
DIM = "dimension"
DIMS = "dimensions"
ROWDIM = "ROWS"
COLDIM = "COLUMNS"
# Range dimensions:
STARTROW = "startRowIndex"
ENDROW = "endRowIndex"
STARTCOL = "startColumnIndex"
ENDCOL = "endColumnIndex"
STARTIDX = "startIndex"
ENDIDX = "endIndex"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autodrive', 'autodrive.formatting']

package_data = \
{'': ['*']}

install_requires = \
['google-api-python-client-stubs>=1.2.0,<2.0.0',
 'google-api-python-client>=2.0.2,<3.0.0',
 'google-auth-httplib2>=0.1.0,<0.2.0',
 'google-auth-oauthlib>=0.4.3,<0.5.0',
 'jsonlines>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'autodrive',
    'version': '0.6.1',
    'description': 'Simple but robust tool for interacting with the Google Drive and Sheets apis via python.',
    'long_description': '# Autodrive\n\nAutodrive is designed to make it as easy as possible to interact with the Google\nDrive and Sheets APIs via Python. It is especially designed to provide as much\nassistance as possible when writing code through hints and autocompletion, as well\nas via thorough type checking and hinting. These features are currently optimized\nfor VSCode, which you can download <a href="https://code.visualstudio.com/">here</a>\nif you wish. They should also work in other Python IDEs.\n\n---\n\n**Documentation:** https://autodrive-py.readthedocs.io/en/latest/\n\n---\n\n## Requirements\n\n---\n\nPython 3.8+\n\n## Installation\n\n---\n\n### Google API Credentials\n\nFollow the steps outlined in the Prerequisites section\n<a href="https://developers.google.com/drive/api/v3/quickstart/python">here</a>.\nDownload and save the `credentials.json` file to the working directory you want to\nuse Autodrive in.\n\n### First Connection\n\nTo test that your credentials provide the expected connection to your Google Drive\naccount, simply instantiate an Autodrive `Drive` instance:\n\n```\nfrom autodrive import Drive\n\ndrive = Drive()\n```\n\nIf your credentials file was saved as `credentials.json`, your browser should\nautomatically open and prompt you to authorize the GCP project you created to\naccess your Google Drive. Click the various Allow prompts it will show you to\ncomplete your first connection. After you see the browser switch to a page\nindicating you can close the process, you should see a `gdrive_token.json` file\nadded to the working directory you saved your `credentials.json` file in. Next time\nyou use an Autodrive element that needs to connect to your Drive, this token will\nbe used and you will not be prompted to authorize access again until it expires.\n\n## Quickstart\n\n---\n\nThe `Drive` class provides methods for finding and creating objects in your Google\nDrive, such as Folders or Sheets.\n\n```\ngsheet = drive.create_gsheet("my-autodrive-gsheet")\n```\n\n### Finding IDs\n\nIf you use `Drive` to search for your Sheets and Folders, you don\'t need to supply the\nGSheet or Folder IDs yourself, but if you know exactly what Sheet you want, then you\ncan directly instantiate a GSheet or folder by pulling the necessary info from the\nobject\'s url.\n\nFor example, if your Sheet\'s url looks like this:\n\n<p>\ndocs.google.com/spreadsheets/d/19k5cT9Klw1CA8Sum-olP7C0JUo6_kMiOAKDEeHPiSr8/edit#gid=0\n</p>\n\nSimply copy/paste the id between `/d/` and `/edit#` as the `gsheet_id`:\n\n```\nfrom autodrive import GSheet\n\ngsheet = GSheet(gsheet_id="19k5cT9Klw1CA8Sum-olP7C0JUo6_kMiOAKDEeHPiSr8")\n```\n\n> **Tabs:** Because Google calls spreadsheets "Sheets", and their api also refers\n> to the individual sub-sheets in a spreadsheet as "Sheets", Autodrive instead\n> refers to them as "Tabs" for clarity.\n\nFor a tab, you can get the `tab_id` from:\n\n<p>\ndocs.google.com/spreadsheets/d/19k5cT9Klw1CA8Sum-olP7C0JUo6_kMiOAKDEeHPiSr8/edit#gid=234276686\n</p>\n\n```\nfrom autodrive import Tab\n\ntab = Tab(\n    gsheet_id="19k5cT9Klw1CA8Sum-olP7C0JUo6_kMiOAKDEeHPiSr8",\n    tab_title="Sheet2",\n    tab_idx=0,\n    tab_id=234276686\n)\n```\n\nFor a folder:\n\n<p>\ndrive.google.com/drive/u/1/folders/1wLx-KMG2jO498xa5ZumB-SEpL-TwczZI\n</p>\n\n```\nfrom autodrive import Folder\n\nfolder = Folder(folder_id="1wLx-KMG2jO498xa5ZumB-SEpL-TwczZI", name="Test Folder")\n```\n\n### Reading and Writing\n\nYou can easily download and write data from a Google Sheet using the `GSheet`,\n`Tab`, or `Range` views.\n\n```\n# Fetches all the data in all cells of the tab:\ntab.get_data()\n\n# Writes 8 cells (2 rows of 4 columns, starting in cell A1) to the tab:\ntab.write_values(\n    [\n        [1, 2, 3, 4],\n        [5, 6, 7, 8],\n    ]\n)\n```\n\n`GSheet` and `Range` have very similar methods, and all of them allow you to read\nand write data to only a specific range in the Google Sheet. See the\n<a href="https://autodrive-py.readthedocs.io/en/latest/">Documentation</a>\nfor more.\n',
    'author': 'Chris Larabee',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/chrislarabee/autodrive',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import Any, Dict

from ..interfaces import BorderFormat, FullRange, Color
from ..dtypes import UserEnteredFmt, UserEnteredVal
from .. import _google_terms as terms


def add_alternating_row_background(
    tab_id: int, rng: FullRange, colors: Color
) -> Dict[str, Any]:
    return {
        "addConditionalFormatRule": {
            "rule": {
                "ranges": [{terms.TAB_ID: tab_id, **dict(rng)}],
                "booleanRule": {
                    "condition": {
                        "type": "CUSTOM_FORMULA",
                        "values": [{str(UserEnteredVal): "=MOD(ROW(), 2)"}],
                    },
                    "format": {"backgroundColor": dict(colors)},
                },
            },
            "index": rng.start_row,
        }
    }


def set_background_color(tab_id: int, rng: FullRange, color: Color):
    return {
        terms.RPT_CELL: {
            terms.RNG: {terms.TAB_ID: tab_id, **dict(rng)},
            terms.FIELDS: str(UserEnteredFmt),
            terms.CELL: {str(UserEnteredFmt): {"backgroundColor": dict(color)}},
        }
    }


def set_border_format(tab_id: int, rng: FullRange, *borders: BorderFormat):
    return {
        terms.RPT_CELL: {
            terms.RNG: {terms.TAB_ID: tab_id, **dict(rng)},
            terms.FIELDS: str(UserEnteredFmt),
            terms.CELL: {
                str(UserEnteredFmt): {
                    "borders": {**{k: v for b in borders for k, v in dict(b).items()}}
                }
            },
        }
    }



from __future__ import annotations

from typing import Any, Dict, Tuple, Sequence, Literal

from . import _google_terms as terms
from .connection import SheetsConnection
from ._view import Component
from .formatting.format_tab import (
    TabCellFormatting,
    TabGridFormatting,
    TabTextFormatting,
)
from .interfaces import AuthConfig, FullRange
from .range import Range
from .dtypes import EffectiveVal, GoogleValueType


class Tab(Component[TabCellFormatting, TabGridFormatting, TabTextFormatting]):
    """
    Provides a connection to a single Google Sheet Tab, its properties, and its
    data.
    """

    def __init__(
        self,
        gsheet_id: str,
        tab_title: str,
        tab_idx: int,
        tab_id: int,
        column_count: int = 26,
        row_count: int = 1000,
        *,
        auth_config: AuthConfig | None = None,
        sheets_conn: SheetsConnection | None = None,
        autoconnect: bool = True,
    ) -> None:
        """

        Args:
            gsheet_id (str): The id string of the target Google Sheet that the Tab
                resides in; can be found in the Google Sheet url.
            tab_title (str): The name of the Tab.
            tab_idx (int): The index (0-based) of the Tab in the Google Sheet.
            tab_id (int): The id of the Tab, can be found in the Google Sheet url
                (after gid=).
            column_count (int, optional): The starting number of columns in the
                Tab, defaults to 26.
            row_count (int, optional): The starting number of rows in the Tab,
                defaults to 1000.
            auth_config (AuthConfig, optional): Optional custom AuthConfig, defaults
                to None.
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.
            autoconnect (bool, optional): If you want to instantiate a Tab without
                immediately checking your authentication credentials and connection
                to the Google Sheets api, set this to False, defaults to True.

        """
        self._tab_id = tab_id
        self._title = tab_title
        self._index = tab_idx
        self._column_count = column_count
        self._row_count = row_count
        super().__init__(
            gsheet_id=gsheet_id,
            gsheet_range=FullRange(
                start_row=0,
                end_row=row_count,
                start_col=0,
                end_col=column_count,
                tab_title=tab_title,
                base0_idxs=True,
            ),
            tab_id=self._tab_id,
            grid_formatting=TabGridFormatting,
            text_formatting=TabTextFormatting,
            cell_formatting=TabCellFormatting,
            auth_config=auth_config,
            sheets_conn=sheets_conn,
            autoconnect=autoconnect,
        )

    @property
    def tab_id(self) -> int:
        """
        Returns:
            int: This Tab's id. Matches the gid of the parent GSheet's url.

        """
        return self._tab_id

    @property
    def format_grid(self) -> TabGridFormatting:
        """
        Returns:
            TabGridFormatting: An object with grid formatting methods.

        """
        return self._format_grid

    @property
    def format_text(self) -> TabTextFormatting:
        """
        Returns:
            TabTextFormatting: An object with text formatting methods.

        """
        return self._format_text

    @property
    def format_cell(self) -> TabCellFormatting:
        """
        Returns:
            TabCellFormatting: An object with cell formatting methods.

        """
        return self._format_cell

    @property
    def title(self) -> str:
        """
        Returns:
            str: The name of this Tab.

        """
        return self._title

    @property
    def index(self) -> int:
        """
        Returns:
            int: The (0-based) index location of this Tab among the other Tabs on
            the parent GSheet.

        """
        return self._index

    @property
    def column_count(self) -> int:
        """
        Returns:
            int: The number of columns in this Tab.

        """
        return self._column_count

    @property
    def row_count(self) -> int:
        """
        Returns:
            int: The number of rows in this Tab.

        """
        return self._row_count

    def fetch(self) -> Tab:
        """
        Gets the latest metadata from the API for this Tab. Re-populates tab
        properties like row and column count.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Returns:
            Tab: This GSheet
        """
        properties = self.conn.get_properties(self._gsheet_id)
        _, sheets = self._parse_properties(properties)
        title, index, tab_id, column_count, row_count = self._unpack_tab_properties(
            sheets[self._index]
        )
        self._title = title
        self._index = index
        self._tab_id = tab_id
        self._column_count = column_count
        self._row_count = row_count
        return self

    @staticmethod
    def _unpack_tab_properties(
        properties: Dict[str, Any]
    ) -> Tuple[str, int, int, int, int]:
        title = str(properties[terms.TAB_NAME])
        index = int(properties[terms.TAB_IDX])
        tab_id = int(properties[terms.TAB_ID])
        column_count = int(properties[terms.GRID_PROPS][terms.COL_CT])
        row_count = int(properties[terms.GRID_PROPS][terms.ROW_CT])
        return title, index, tab_id, column_count, row_count

    @classmethod
    def from_properties(
        cls,
        gsheet_id: str,
        properties: Dict[str, Any],
        auth_config: AuthConfig | None = None,
        sheets_conn: SheetsConnection | None = None,
        autoconnect: bool = True,
    ) -> Tab:
        """
        Generates a Tab assigned to the passed gsheet_id with the passed tab
        properties dictionary from a SheetsConnection.get_properties call.

        Unless you have a special use-case, it is probably more trouble than it's
        worth to try to instantiate a Tab with this method, as it is designed for
        use by other Autodrive objects.

        Args:
            gsheet_id (str): The id of the parent GSheet.
            properties (Dict[str, Any]): A properties dictionary, which must
                contain index, SheetId, title, and gridProperties keys. The
                gridProperties must be a dictionary containing columnCount and
                rowCount keys.
            auth_config (AuthConfig, optional): Optional custom AuthConfig,
                defaults to None.
            sheets_conn (SheetsConnection, optional): Optional manually created
                SheetsConnection, defaults to None.
            autoconnect (bool, optional): If you want to instantiate a Tab without
                immediately checking your authentication credentials and connection
                to the Google Sheets api, set this to False, defaults to True.

        Returns:
            Tab: A Tab with the values from the passed properties dictionary.

        """
        title, index, tab_id, column_count, row_count = cls._unpack_tab_properties(
            properties
        )
        return Tab(
            gsheet_id,
            tab_title=title,
            tab_idx=index,
            tab_id=tab_id,
            column_count=column_count,
            row_count=row_count,
            auth_config=auth_config,
            sheets_conn=sheets_conn,
            autoconnect=autoconnect,
        )

    def full_range(self) -> FullRange:
        """
        Generates a FullRange object corresponding to the full range of the Tab.

        Returns:
            FullRange: A FullRange from A1:the end of the Tab.

        """
        return FullRange(
            start_row=0,
            end_row=self._row_count,
            start_col=0,
            end_col=self._column_count,
            base0_idxs=True,
        )

    def get_data(
        self,
        rng: FullRange | str | None = None,
        value_type: GoogleValueType = EffectiveVal,
    ) -> Tab:
        """
        Gets the data from the cells of this Tab.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Args:
            rng (FullRange | HalfRange, optional): An optional range value, to
                specify a subset of the Tab's values to get, defaults to None,
                which fetches all values in the Tab.
            value_type (GoogleValueType, optional): Allows you to toggle the
                type of the values returned by the Google Sheets API. See the
                :mod:`dtypes <autodrive.dtypes>` documentation for more info on
                the different GoogleValueTypes.

        Returns:
            Tab: This Tab.

        """
        rng = self.ensure_full_range(self.full_range(), rng)
        if not rng.tab_title:
            rng.tab_title = self._title
        self._values, self._formats = self._get_data(
            self._gsheet_id, str(rng), value_type
        )
        return self

    def write_values(
        self,
        data: Sequence[Sequence[Any] | Dict[str, Any]],
        rng: FullRange | str | None = None,
        mode: Literal["write", "w", "append", "a"] = "write",
    ) -> Tab:
        """
        Adds a request to write data. Tab.commit () to commit the requests.

        Args:
            data (Sequence[Sequence[Any] | Dict[str, Any]]): The data to write.
                Each sequence or dictionary in the passed data is a row, with
                each value in that sub-iterable being a column. Dictionary keys
                will be used as a header row in the written data.
            rng (FullRange, optional): A specific range to write to,
                starting with the top-left-most cell in the range, defaults to None,
                which will write to the top-left-most cell of the Tab.
            mode (Literal, optional): Whether to append the data after any
                populated rows already present in the Tab or to write to the
                passed rng. Overrides rng if specified. Defaults to "write".

        Returns:
            Tab: This Tab.

        """
        rng = self.ensure_full_range(self.full_range(), rng)
        rng_dict = rng.to_dict() if mode in ["write", "w"] else None
        self._write_values(data, self._tab_id, rng_dict)
        return self

    @classmethod
    def new_tab_request(
        cls,
        tab_title: str,
        tab_id: int | None = None,
        tab_idx: int | None = None,
        num_rows: int = 1000,
        num_cols: int = 26,
    ) -> Dict[str, Any]:
        """
        Creates a dictionary request to create a new tab in a Google Sheet.

        Args:
            tab_title (str): The name of the tab.
            tab_id (int, optional): The desired id of the tab, which cannot already
                exist in the Google Sheet, defaults to None, which will allow the
                Google Sheet to generate the tab_id.
            tab_idx (int, optional): The (0-based) index to create the tab at,
                defaults to None, meaning the tab will be created as the last tab
                in the Google Sheet.
            num_rows (int, optional): The starting number of rows in the new tab,
                defaults to 1000.
            num_cols (int, optional): The starting number of columns in the new
                tab, defaults to 26.

        Returns:
            Dict[str, Any]: A dictionary ready to be passed as a request via a
            request commit.

        """
        props: Dict[str, Any] = {
            terms.TAB_NAME: tab_title,
            terms.GRID_PROPS: {terms.COL_CT: num_cols, terms.ROW_CT: num_rows},
        }
        if tab_id is not None:
            props[terms.TAB_ID] = tab_id
        if tab_idx is not None:
            props[terms.TAB_IDX] = tab_idx
        result = {terms.ADDTAB: {terms.TAB_PROPS: props}}
        return result

    def gen_add_tab_request(self) -> Dict[str, Any]:
        """
        Generates a new tab request dictionary for this Tab. Useful when you have
        manually instantiated a Tab object instead of fetching it and want to add
        it to the parent GSheet.

        Returns:
            Dict[str, Any]: A dictionary ready to be passed as a request via a
            request commit.

        """
        return self.new_tab_request(
            self._title, self._tab_id, self._index, self._row_count, self._column_count
        )

    def create(self) -> Tab:
        """
        Convenience method for generating a new tab request based on this Tab and
        immediately committing it, thus adding it to the parent Google Sheet.

        .. note::

            This method will cause a request to be posted to the relevant Google
            API immediately.

        Returns:
            Tab: This Tab.

        """
        req = self.new_tab_request(
            self.title, self.tab_id, self.index, self.row_count, self.column_count
        )
        self._requests.append(req)
        self.commit()
        return self

    def gen_range(self, rng: FullRange) -> Range:
        """
        Convenience method for generating a new Range object in this Tab.

        Args:
            rng (FullRange): The desired FullRange of the new Range object.

        Returns:
            Range: The newly generated Range.
        """
        return Range(
            rng, self._gsheet_id, self._title, self._tab_id, sheets_conn=self._conn
        )



from __future__ import annotations

from ._core import GoogleDtype


class String(metaclass=GoogleDtype):
    """
    String datatype, covers all non-numeric/date text that isn't a formula.
    """

    python_type = str
    type_key = "stringValue"

    @classmethod
    def parse(cls, value: str) -> str:
        """
        Converts a string into a string.

        Args:
            value (str): Any string.

        Returns:
            str: The passed value as a string.

        """
        return cls.python_type(value)


class Formula(metaclass=GoogleDtype):
    """
    Formula datatype, essentially a string, but begins with =.
    """

    python_type = str
    type_key = "formulaValue"

    @classmethod
    def parse(cls, value: str) -> str:
        """
        Converts a string into a formula (string).

        Args:
            value (str): Any string.

        Returns:
            str: The passed value as a string.

        """
        return cls.python_type(value)


class Number(metaclass=GoogleDtype):
    """
    Number datatype, covers both floats and integers, though internally Number is
    treated as a float so as to not lose significant digits.
    """

    python_type = float
    type_key = "numberValue"

    @classmethod
    def parse(cls, value: str) -> float | int:
        """
        Converts a string into an integer (if it has no decimal point) or float.

        Args:
            value (str): Any numeric string.

        Returns:
            float | int: The passed value as an integer or float.

        """
        if value.isnumeric():
            return int(value)
        else:
            return cls.python_type(value)


class Boolean(metaclass=GoogleDtype):
    """
    Boolean datatype, appears in Google Sheets as FALSE or TRUE.
    """

    python_type = bool
    type_key = "boolValue"

    @classmethod
    def parse(cls, value: str) -> bool:
        """
        Converts a string into a boolean.

        Args:
            value (str): Any string.

        Returns:
            bool: False if the string is some variation of FALSE, otherwise True for
            all other strings.

        """
        if value.lower() == "false":
            return False
        else:
            return True


class GoogleValueType(type):
    """
    Metaclass for the three different ways values are stored in Google Sheets.
    """

    value_key: str
    """The API key for the GoogleValueType."""
    has_dtype: bool = True
    """Whether dtype information can be extracted from the GoogleValueType."""

    def __str__(cls) -> str:
        return cls.value_key


class UserEnteredVal(metaclass=GoogleValueType):
    """
    UserEnteredVal is the value as entered by the user, without any calculations
    applied to it. It is always a string, but is accompanied by metadata that
    indicates what datatype the value is.
    """

    value_key = "userEnteredValue"


class EffectiveVal(metaclass=GoogleValueType):
    """
    EffectiveVal is the value as displayed in Google Sheets, and is appropriately
    typed when read from the api. So if the formula "=A1+A2" would equal 3, then
    the EffectiveVal of that cell is 3.
    """

    value_key = "effectiveValue"


class FormattedVal(metaclass=GoogleValueType):
    """
    The untyped string value of the cell as displayed in Google Sheets. Essentially
    equivalent to EffectiveVal, but without appropriate typing.
    """

    value_key = "formattedValue"
    has_dtype = False


GOOGLE_VAL_TYPES = (UserEnteredVal, EffectiveVal, FormattedVal)
"""A tuple of UserEnteredVal, EffectiveVal, and FormattedVal google value types."""

GOOGLE_DTYPES = (String, Formula, Number, Boolean)
"""A tuple of String, Formula, Number, and Boolean google data types."""

TYPE_MAP = {
    str: String,
    int: Number,
    float: Number,
    bool: Boolean,
}
"""Dictionary mapping python data types to corresponding google data types."""

REV_TYPE_MAP = {
    String: String.python_type,
    Formula: Formula.python_type,
    Number: Number.python_type,
    Boolean: Boolean.python_type,
}
"""Dictionary mapping google data types to corresponding python types."""


class _Property:
    """
    Abstract base class for various properties that will be instantiated as
    attributes on this module.
    """

    def __init__(self, property: str) -> None:
        self.prop = property

    def __str__(self) -> str:
        return self.prop


UserEnteredFmt = _Property("userEnteredFormat")
"""The formatting properties the user as applied to the cell. """

EffectiveFmt = _Property("effectiveFormat")
"""
The final formatting information about a cell and its values, includes the 
effects of conditional formatting and the like.
"""

DefaultFmt = _Property("defaultFormat")
"""The default formatting properties of a cell, dictated by the settings of the tab."""


class BorderStyle(_Property):
    """
    A property describing the style of border to apply to part of a cell.
    """

    def __init__(self, style: str) -> None:
        """
        Args:
            style (str): The name of the style.
        """
        super().__init__(style)


BorderSolid = BorderStyle("SOLID")
"""The default border style, a thin, solid line."""

BorderSolidMedium = BorderStyle("SOLID_MEDIUM")
"""Same as BorderSolid, but slightly thicker."""

BorderSolidThick = BorderStyle("SOLID_THICK")
"""Same as BorderSolid, but much thicker."""

BorderDashed = BorderStyle("DASHED")
"""A thin line comprised of dashes."""

BorderDotted = BorderStyle("DOTTED")
"""A thin line comprised of dots."""

BorderDoubleLine = BorderStyle("DOUBLE")
"""A set of two parallel lines."""


class BorderSide(_Property):
    """
    A property describing which side of a cell to apply border properties to.
    """

    def __init__(self, side: str) -> None:
        """
        Args:
            side (str): The name of the side.
        """
        super().__init__(side)


BorderLeft = BorderSide("left")
"""The border for the left side of a cell."""

BorderRight = BorderSide("right")
"""The border for the right side of a cell."""

BorderTop = BorderSide("top")
"""The border for the top side of a cell."""

BorderBottom = BorderSide("bottom")
"""The border for the bottom side of a cell."""

BorderSides = (BorderLeft, BorderRight, BorderTop, BorderBottom)
"""Convenience reference for all BorderSide objects."""


class VerticalAlign(_Property):
    """
    A property describing vertical text alignment.
    """

    def __init__(self, align_str: str) -> None:
        super().__init__(align_str)


class HorizontalAlign(_Property):
    """
    A property describing horizontal text alignment.
    """

    def __init__(self, align_str: str) -> None:
        super().__init__(align_str)


AlignTop = VerticalAlign("TOP")
"""Align text to the top of the cell(s)."""

AlignMiddle = VerticalAlign("MIDDLE")
"""Align text to the middle of the cell(s)."""

AlignBottom = VerticalAlign("BOTTOM")
"""Align text to the middle of the cell(s)."""

AlignLeft = HorizontalAlign("LEFT")
"""Align text to the left of the cell(s)."""

AlignCenter = HorizontalAlign("CENTER")
"""Align text to the center of the cell(s)."""

AlignRight = HorizontalAlign("RIGHT")
"""Align text to the right of the cell(s)."""



from __future__ import annotations

import re
import string
from pathlib import Path
from typing import Any, Dict, Iterator, List, Mapping, Optional, Tuple, TypeVar

from .dtypes import UserEnteredFmt, BorderStyle, BorderSide, BorderSolid
from . import _google_terms as terms

DEFAULT_TOKEN = "gdrive_token.json"
"""Filepath for the token json file. Default="gdrive_token.json". """
DEFAULT_CREDS = "credentials.json"
"""Filepath for the credentials json file. Default="credentials.json" """


_T = TypeVar("_T")


class ParseRangeError(Exception):
    """
    Error raised when an invalid range string is passed to a FullRange.
    """

    def __init__(self, rng: str, msg_addon: str | None = None, *args: object) -> None:
        """
        Args:
            rng (str): The original range string.
            msg_addon (str, optional): An addendum to the error message. Defaults to
                None.
        """
        msg = f"{rng} is not a valid range.{' ' + msg_addon if msg_addon else ''}"
        super().__init__(msg, *args)


class AuthConfig:
    """
    Optional custom configuration for Autodrive's authentication processes using
    your Google api credentials.
    """

    def __init__(
        self,
        secrets_config: Dict[str, Any] | None = None,
        token_filepath: str | Path = DEFAULT_TOKEN,
        creds_filepath: str | Path = DEFAULT_CREDS,
    ) -> None:
        """
        Args:
            secrets_config (Dict[str, Any], optional): The dictionary of
                configuration used by a google.oath2.credentials Credentials
                object. Generally only useful if for some reason you are reading
                in your own token file, defaults to None.
            token_filepath (str | Path, optional): The filepath to your gdrive_token
                json file. This doesn't have to exist at time of authentication,
                and will be saved to this path when the authorization flow completes,
                defaults to DEFAULT_TOKEN, which is "gdrive_token.json" in your
                cwd.
            creds_filepath (str | Path, optional): The filepath to your api
                credentials json file. This file *does* need to exist at time of
                authentication, unless you pass a secrets_config dictionary,
                defaults to DEFAULT_CREDS, which is "credentials.json" from your
                cwd.

        """
        self.secrets_config = secrets_config
        self._token_filepath = Path(token_filepath)
        self._creds_filepath = Path(creds_filepath)

    @property
    def token_filepath(self) -> Path:
        """
        Returns:
            Path: The Path to your token json file.

        """
        return self._token_filepath

    @property
    def creds_filepath(self) -> Path:
        """
        Returns:
            Path: The Path to your credentials json file.

        """
        return self._creds_filepath


class _Interface(Mapping[str, _T]):
    """
    Base class for all Interfaces.
    """

    def to_dict(self) -> Dict[str, _T]:
        return {}

    def __iter__(self) -> Iterator[str]:
        return iter(self.to_dict())

    def __getitem__(self, k: str) -> _T:
        return self.to_dict()[k]

    def __len__(self) -> int:
        return len(self.to_dict())


class _RangeInterface(_Interface[int]):
    """
    Underlying class for HAlf and Full Range Interfaces.
    """

    def __init__(self, tab_title: str | None = None) -> None:
        self.tab_title = tab_title

    def to_dict(self) -> Dict[str, int]:
        return {}

    @classmethod
    def _construct_range_str(
        cls,
        start_row: int = 0,
        start_col: int = 0,
        end_row: Optional[int] = None,
        end_col: Optional[int] = None,
    ) -> str:
        """
        Reconstructs a standardized range string (Sheet1!A1:B3) from the passed
        properties.

        Args:
            start_row (int, optional): 0-based row index for the first
                row in the range, defaults to 0.
            start_col (int, optional): 0-based column index for the first
                column in the range, defaults to 0.
            end_row (int, optional): 0-based row index for the row after
                the range, defaults to None.
            end_col (int, optional): 0-based column index for the column
                after the range, defaults to None.

        Returns:
            str: A string in the format (tab_name!start_cell:end_cell)

        """
        start_letter = ""
        end_letter = ""
        end_row_int = end_row + 1 if end_row else end_row
        end_col_int = end_col if end_col is not None else start_col
        start_letter = cls._convert_col_idx_to_alpha(start_col)
        end_letter = cls._convert_col_idx_to_alpha(end_col_int)
        start_int = start_row + 1
        if end_col or end_row:
            end_part = f":{end_letter}{end_row_int or ''}"
        else:
            end_part = ""
        return f"{start_letter}{start_int}{end_part}"

    @staticmethod
    def _parse_range_str(rng: str) -> Tuple[Optional[str], str, Optional[str]]:
        """
        Parses a range string (Sheet1:A1:B3) into its component groups ("Sheet1",
        "A1", "B3").

        Args:
            rng (str): The range string, which must at least be one cell coordinate
                (e.g. A1), which will be returned as the start cell.

        Returns:
            Tuple[Optional[str], str, Optional[str]]: The sheet title (if present),
                a start cell, and an end cell (if present).

        Raises:
            ParseRangeError: If the range is invalid.

        """
        result = re.match(r"(?:(.*)!)?([A-Z]+\d+)(?::([A-Z]*\d*))?", rng)
        if result:
            grps = result.groups()
            return grps  # type: ignore
        else:
            raise ParseRangeError(rng)

    @staticmethod
    def _parse_cell_str(cell_str: str) -> Tuple[str, Optional[str]]:
        """
        Parses an individual cell string (A1) into its component strings ("A", "1")

        Args:
            cell_str (str: str): The cell string, which must at least have a
                column letter.

        Returns:
            Tuple[str, Optional[str]]: The column letter and the row number.

        Raises:
            ParseRangeError: If the range is invalid.

        """
        result = re.match(r"([A-Z]+)(\d+)?", cell_str)
        if result:
            grps = result.groups()
            return grps  # type: ignore
        else:
            raise ParseRangeError(cell_str)

    @classmethod
    def _convert_cell_str_to_coord(cls, cell_str: str) -> Tuple[int, Optional[int]]:
        """
        Parses the passed cell string (e.g. A1) and returns its numeric (0-based)
        indices (e.g. 0, 0)

        Args:
            cell_str (str): The cell string.

        Returns:
            Tuple[int, Optional[int]]: The column index and the row index (if
            present).

        """
        col, row = cls._parse_cell_str(cell_str)
        col_idx = cls._convert_alpha_col_to_idx(col)
        row_idx = int(row) - 1 if row else None
        return col_idx, row_idx

    @staticmethod
    def _convert_alpha_col_to_idx(alpha_col: str) -> int:
        """
        Converts a string column identifier (A, B, C, ...Z, AA, AAA, etc) into a
        (0 based) numeric index.

        Args:
            alpha_col (str): A letter or set of letters.

        Returns:
            int: The numeric representation of the alpha_col's index.

        """
        values: List[int] = []
        for i, a in enumerate(alpha_col, start=1):  # type: ignore
            base_idx = string.ascii_uppercase.index(a) + 1  # type: ignore
            remainder = len(alpha_col[i:])
            values.append(26 ** remainder * base_idx)
        return sum(values) - 1

    @staticmethod
    def _convert_col_idx_to_alpha(idx: int) -> str:
        """
        Converts a (0-based) index into a string column identifier.

        Args:
            idx (int): A (0-based) index value.

        Returns:
            str: The string representation of the idx's position.

        """
        chars: List[str] = []
        col_num = idx + 1
        while col_num > 0:
            remainder = col_num % 26
            if remainder == 0:
                remainder = 26
            col_letter = chr(ord("A") + remainder - 1)
            chars.append(col_letter)
            col_num = int((col_num - 1) / 26)
        chars.reverse()
        return "".join(chars)

    @classmethod
    def _parse_idx(cls, idx: str | int, base0_idxs: bool = False) -> int:
        """
        Extracted logic for standardizing Half and Full start/end indices. Allows
        the end user to supply strings or integers as desired to the Ranges, and
        also allows them to use 0-based indices or not.

        Args:
            idx (str | int): An index (optionally 0-based) or a string
                column representation.
            base0_idxs (bool, optional): Set to True if passing 0-based indices,
                otherwise will assume it needs to adjust them. defaults to False.
            mod (int, optional): The value by which to adjust non-0-based indices,
                defaults to 1.

        Returns:
            Optional[int]: A 0-based numeric representation of the passed index.

        """
        adj = 0 if base0_idxs else -1
        if isinstance(idx, str):
            result = cls._convert_alpha_col_to_idx(idx)
        else:
            result = idx + adj
        return result


class HalfRange(_RangeInterface):
    """
    A partial range used in requests to the Google Sheets API when the axis of the
    range is obvious from context (such as when inserting rows or columns).

    Note that HalfRange parses your inputs into Google Sheets API row/column
    indices, which are 0-based, so if you call :attr:`start_idx` or :attr:`end_idx`
    expect them to be 1 lower than the value you passed.
    """

    def __init__(
        self,
        start_idx: str | int | None = None,
        end_idx: str | int | None = None,
        *,
        tab_title: str | None = None,
        base0_idxs: bool = False,
        column: bool = False,
    ) -> None:
        """

        Args:
            start_idx (str | int, optional): The first column/row in the range,
                assumed to not be 0-based if it's an integer (row), defaults to
                None.
            end_idx (str | int, optional): The last column/row in the range,
                assumed to not be 0-based if it's an integer (row), defaults to
                None.
            tab_title (str, optional): The name of the tab this range resides in,
                defaults to None.
            base0_idxs (bool, optional): Set to True if you'd like to pass 0-based
                indices to start_idx and end_idx params, above, defaults to False.
            column (bool, optional): Set to True if this HalfRange is a column range
                and you supplied the column indexes as integers, defaults to False.

        """
        super().__init__(tab_title)
        if start_idx is not None:
            self.start_idx = self._parse_idx(start_idx, base0_idxs)
        elif end_idx is not None:
            self.start_idx = 0
        if end_idx is not None:
            self.end_idx = self._parse_idx(end_idx, base0_idxs) if end_idx else None
        elif self.start_idx is not None:
            self.end_idx = self.start_idx
        self.column = column
        if isinstance(start_idx, str) or isinstance(end_idx, str):
            self.column = True

    def to_dict(self) -> Dict[str, int]:
        """
        The Google Sheets api ranges are end-value exclusive, so this method will
        produce a dictionary with an endIndex value 1 higher than the FullRange's
        attribute.

        Returns:
            Dict[str, int]: Outputs the HalfRange as a dictionary of properties
            usable in generating an api request to affect the target range of cells.

        """
        result: Dict[str, int] = {}
        # All of these must be is not None because any of them can be 0:
        if self.start_idx is not None:
            result[terms.STARTIDX] = self.start_idx
        if self.end_idx is not None:
            result[terms.ENDIDX] = self.end_idx + 1
        return result

    def __str__(self) -> str:
        if self.column:
            rng = self._construct_range_str(
                start_col=self.start_idx, end_col=self.end_idx
            )
        else:
            raise AttributeError(
                f"{self}.column indicator is False. Cannot convert row HalfRanges to "
                "strings."
            )
        return f"{self.tab_title or ''}{'!' if self.tab_title else ''}{rng}"


class FullRange(_RangeInterface):
    """
    A complete Google Sheets range (indicating start and end row and column as
    well as at least an end column, if not an end row).

    Note that FullRange parses your inputs into Google Sheets API row/column
    indices, which are 0-based, so if you call :attr:`start_row`, :attr:`end_row`,
    :attr:`start_col`, or :attr:`end_col`, expect them to be 1 lower than the
    value you passed.
    """

    def __init__(
        self,
        range_str: str | None = None,
        *,
        start_row: int | None = None,
        end_row: int | None = None,
        start_col: int | str | None = None,
        end_col: int | str | None = None,
        base0_idxs: bool = False,
        tab_title: str | None = None,
    ) -> None:
        """

        Args:
            range_str (str, optional): A range string (e.g. Sheet1!A1:B3, A1:B3, A1).
            start_row (int, optional): The first row in the range, assumed to not
                be 0-based, defaults to None.
            end_row (int, optional): The last row in the range, assumed to not be
                0-based, defaults to None.
            start_col (int | str, optional): The first column in the range, assumed
                to not be 0-based if it's an integer, defaults to None.
            end_col (int | str, optional): The last column in the range, assumed to
                not be 0-based if it's an integer, defaults to None.
            base0_idxs (bool, optional): Set to True if you'd like to pass 0-based
                indices to start/end_row and start/end_col, defaults to False.
            tab_title (str, optional): The name of the tab this range resides in,
                defaults to None.

        """
        super().__init__(tab_title)
        self._single_cell = False
        if range_str:
            title, start, end = self._parse_range_str(range_str)
            if title and not self.tab_title:
                self.tab_title = title
            cstart, rstart = self._convert_cell_str_to_coord(start)
            if rstart is None:
                raise ParseRangeError(range_str)
            if end:
                cend, rend = self._convert_cell_str_to_coord(end)
            else:
                cend = cstart + 1
                rend = rstart + 1
                self._single_cell = True
            self.start_row = rstart
            self.end_row = rend
            self.start_col = cstart
            self.end_col = cend
        else:
            self.start_row = self._parse_idx(start_row, base0_idxs) if start_row else 0
            self.start_col = self._parse_idx(start_col, base0_idxs) if start_col else 0
            if end_row is None:
                self.end_row = self.start_row
            else:
                self.end_row = self._parse_idx(end_row, base0_idxs)
            if end_col is None:
                self.end_col = self.start_col
            else:
                self.end_col = self._parse_idx(end_col, base0_idxs)
            if end_row is None and end_col is None:
                self._single_cell = True

    @property
    def row_range(self) -> HalfRange:
        """
        Returns:
            HalfRange: The FullRange's row range as a HalfRange.

        """
        return HalfRange(
            self.start_row,
            self.end_row,
            tab_title=self.tab_title,
            base0_idxs=True,
        )

    @property
    def col_range(self) -> HalfRange:
        """
        Returns:
            HalfRange: The FullRange's column range as a HalfRange.

        """
        return HalfRange(
            self.start_col,
            self.end_col,
            tab_title=self.tab_title,
            base0_idxs=True,
            column=True,
        )

    def __str__(self) -> str:
        if self._single_cell:
            rng = self._construct_range_str(self.start_row, self.start_col)
        else:
            rng = self._construct_range_str(
                self.start_row, self.start_col, self.end_row, self.end_col
            )
        return f"{self.tab_title or ''}{'!' if self.tab_title else ''}{rng}"

    def to_dict(self) -> Dict[str, int]:
        """
        The Google Sheets api ranges are end-value exclusive, so this method will
        produce a dictionary with endRowIndex and endColumnIndex values 1 higher
        than the FullRange's attribute.

        Returns:
            Dict[str, int]: Outputs the FullRange as a dictionary of properties
            usable in generating an api request to affect the target range of cells.

        """
        result: Dict[str, int] = {}
        # All of these must be is not None because any of them can be 0:
        if self.start_row is not None:
            result[terms.STARTROW] = self.start_row
        if self.end_row is not None:
            result[terms.ENDROW] = self.end_row + 1
        if self.start_col is not None:
            result[terms.STARTCOL] = self.start_col
        if self.end_col is not None:
            result[terms.ENDCOL] = self.end_col + 1
        return result


class BorderFormat(_Interface[Any]):
    """
    The settings for the border of a single side of a cell.
    """

    def __init__(
        self,
        side: BorderSide,
        color: Color | None = None,
        style: BorderStyle | None = None,
    ) -> None:
        """

        Args:
            side (BorderSide): The BorderSide to apply settings to.
            color (Color, optional): The color of the side of the border.
                Defaults to None, for a black border.
            style (BorderStyle, optional): The BorderStyle of the side of the
                border. Defaults to None, for the default border style.
        """
        self.side = side
        self.color = color if color else Color(0, 0, 0)
        self.style = style if style else BorderSolid

    def to_dict(self) -> Dict[str, Any]:
        """
        Returns:
            Dict[str, Any]: Outputs the BorderFormat as a dictionary of properties
            usable in generating an api request to affect cell border properties.

        """
        return {
            str(self.side): {
                "color": dict(self.color),
                "style": str(self.style),
            }
        }


class Color(_Interface[float]):
    """
    An RGBA color value.
    """

    def __init__(
        self,
        red: int | float = 0,
        green: int | float = 0,
        blue: int | float = 0,
        alpha: int | float = 100,
    ) -> None:
        """

        Args:
            red (int | float, optional): Either an integer from 0 to 255 (from
                which a float value will be calculated), or the float representation
                of same. defaults to 0.
            green (int | float, optional): Either an integer from 0 to 255 (from
                which a float value will be calculated), or the float representation
                of same. defaults to 0.
            blue (int | float, optional): Either an integer from 0 to 255 (from
                which a float value will be calculated), or the float representation
                of same. defaults to 0.
            alpha (int, optional): Either an integer from 0 to 100 (from which a
                float value will be calculated), or the float representation of
                same. defaults to 100.

        """
        self.red = self._ensure_valid_input(red, 255)
        self.green = self._ensure_valid_input(green, 255)
        self.blue = self._ensure_valid_input(blue, 255)
        self.alpha = self._ensure_valid_input(alpha, 100)

    @staticmethod
    def _ensure_valid_input(input: int | float, intmax: int = 100) -> float:
        """
        Ensures the passed input is between 0 and 1, or the passed maximum int
        value if input is an integer. Any value below 0 will be set to 0, and
        any value above 1 or the intmax will be set to 1 or the intmax.

        Args:
            input (int | float): Any integer or float.
            intmax (int, optional): Maximum allowed value of integers. Defaults
                to 100.

        Returns:
            float: A float representation of the input. Integers will be divided
                by the intmax to calculate the return value.
        """
        if input < 0:
            input = 0
        elif isinstance(input, float):
            if input > 1:
                input = 1.0
        else:
            if input > intmax:
                input = intmax
            input = input / intmax
        return input

    def to_dict(self) -> Dict[str, float]:
        """
        Returns:
            Dict[str, float]: Outputs the Color as a dictionary of properties
            usable in generating an api request to affect the color of text or
            cell background.

        """
        return {
            "red": self.red,
            "green": self.green,
            "blue": self.blue,
            "alpha": self.alpha,
        }

    @classmethod
    def from_hex(cls, hex_code: str, alpha: int | float = 100) -> Color:
        """
        Instantiates a Color object from the supplied hex code.

        Args:
            hex_code (str): A hexadecimal color code.
            alpha (int | float, optional): Optional alpha parameter (not included in hex
                codes). Defaults to 100.

        Returns:
            Color: The new Color object.
        """
        hex_code = hex_code.replace("#", "")
        r, g, b = tuple(int(hex_code[i : i + 2], 16) for i in (0, 2, 4))
        return Color(r, g, b, alpha)


class Format(_Interface[Any]):
    """
    Underlying class for text/number formats.
    """

    def __init__(self, format_key: str) -> None:
        """
        Args:
            format_key (str): The format key as dictated by the Google Sheets api.

        """
        self._format_key = format_key

    @property
    def format_key(self) -> str:
        """
        Returns:
            str: The format's Google Sheets api key, used to generate the
            userEnteredFormat value.

        """
        return self._format_key

    def __str__(self) -> str:
        return f"{UserEnteredFmt}({self._format_key})"

    def _fmt_contents(self) -> Dict[str, Any]:
        return {}

    def to_dict(self) -> Dict[str, Any]:
        """
        Returns:
            Dict[str, Any]: Outputs the Format as a dictionary of properties
            usable in generating an api request to affect the text/number format
            of one or more cells.

        """
        result = self._fmt_contents()
        return {str(UserEnteredFmt): {self._format_key: result}}


class TextFormat(Format):
    """
    Provides parameters available in updating the text formatting of a cell.
    """

    def __init__(
        self,
        *,
        font: str | None = None,
        color: Color | None = None,
        font_size: int | None = None,
        bold: bool | None = None,
        italic: bool | None = None,
        underline: bool | None = None,
        strikethrough: bool | None = None,
    ) -> None:
        """
        Args:
            font (str, optional): The name of a font to change to, defaults to None.
            color (Color, optional): The Color properties to apply to the text,
                defaults to None.
            font_size (int, optional): The size to apply to the text, defaults to
                None.
            bold (bool, optional): Whether to turn bold on (True) or off (False),
                defaults to None, for unchanged.
            italic (bool, optional): Whether to turn italic on (True) or off (False),
                defaults to None, for unchanged.
            underline (bool, optional): Whether to turn underline on (True) or off
                (False), defaults to None, for unchanged.
            strikethrough (bool, optional): Whether to turn strikethrough on (True)
                or off (False), defaults to None, for unchanged.

        """
        super().__init__("textFormat")
        self.font = font
        self.color = color
        self.font_size = font_size
        self.bold = bold
        self.italic = italic
        self.underline = underline
        self.strikethrough = strikethrough

    def _fmt_contents(self) -> Dict[str, bool | str | int | Dict[str, float]]:
        """
        Returns:
            Dict[str, bool | str | int | Dict[str, float]]: The TextFormat as a
            request-ready dictionary.

        """
        result: Dict[str, bool | str | int | Dict[str, float]] = {}
        if self.bold is not None:
            result["bold"] = self.bold
        if self.italic is not None:
            result["italic"] = self.italic
        if self.underline is not None:
            result["underline"] = self.underline
        if self.strikethrough is not None:
            result["strikethrough"] = self.strikethrough
        if self.font:
            result["fontFamily"] = self.font
        if self.color:
            result["foregroundColor"] = self.color.to_dict()
        if self.font_size is not None:
            result["fontSize"] = self.font_size
        return result


class NumericFormat(Format):
    """
    A Google Sheet Number Format, which the value of the cell (s) will be parsed
    into.
    """

    def __init__(self, pattern: str = "") -> None:
        """
        Args:
            pattern (str, optional): A pattern valid as a Google Sheet Number
                Format, defaults to "", for automatic number format.

        """
        super().__init__("numberFormat")
        self.type = "NUMBER"
        self.pattern = pattern

    def _fmt_contents(self) -> Dict[str, str]:
        """
        Returns:
            Dict[str, str]: The NumericFormat as a request-ready dictionary.

        """
        return {"type": self.type, "pattern": self.pattern}


AutomaticFormat = NumericFormat()
"""Corresponds to the Automatic ``1000.12`` format."""

NumberFormat = NumericFormat("#,##0.00")
"""Corresponds to the Number ``1,000.12`` format."""

AccountingFormat = NumericFormat('_($* #,##0.00_);_($* (#,##0.00);_($* "-"??_);_(@_)')
"""Corresponds to the Accounting ``$ (1,000.12)`` format."""

PercentFormat = NumericFormat("0.00%")
"""Corresponds to the Percent ``10.12%`` format."""

ScientificFormat = NumericFormat("0.00E+00")
"""Corresponds to the Scientific ``1.01E+03`` format."""

FinancialFormat = NumericFormat("#,##0.00;(#,##0.00)")
"""Corresponds to the Financial ``(1,000.12)`` format."""

CurrencyFormat = NumericFormat('"$"#,##0.00')
"""Corresponds to the Currency ``$1,000.12`` format."""

CurRoundFormat = NumericFormat('"$"#,##0')
"""Corresponds to the Currency (rounded) ``$1,000`` format."""

DateFormat = NumericFormat("M/d/yyyy")
"""Corresponds to the Date ``9/26/2008`` format."""

TimeFormat = NumericFormat("h:mm:ss am/pm")
"""Corresponds to the Time ``3:59:00 PM`` format."""

DatetimeFormat = NumericFormat("M/d/yyyy H:mm:ss")
"""Corresponds to the Date time ``9/26/2008 15:59:00`` format."""

DurationFormat = NumericFormat("[h]:mm:ss")
"""Corresponds to the Duration ``24:01:00`` format."""



from .format_rng import RangeCellFormatting, RangeGridFormatting, RangeTextFormatting
from .format_tab import TabCellFormatting, TabGridFormatting, TabTextFormatting

__all__ = [
    "RangeCellFormatting",
    "RangeGridFormatting",
    "RangeTextFormatting",
    "TabCellFormatting",
    "TabGridFormatting",
    "TabTextFormatting",
]
